/* $Id: piim_r.cpp 358 2020-03-16 23:03:27Z pjohnson $
  ---------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2009. "Inference of Microbial
    Recombination Rates from Metagenomic Data."  PLoS Genetics (in
    press).
 
  Copyright 2009 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Estimates recombination rate from metagenomic-style data in XML
  format conforming to piim_r.dtd
  ---------------------------------------------------------------------------
*/

#include "mle.hpp"
#include "util.hpp"
#include "util_disklist.hpp"
#include "piim_data.hpp"
#include "freqspec.hpp"
#include <fstream>
#include <sstream>
#include <iomanip>
#include <memory>
#ifdef SYSTEM_MISSING_TR1
#include <map>
#else
#include <tr1/unordered_map>
#endif
#include <gsl/gsl_sf_gamma.h>

/*---------------------------------------------------------------------------*/

enum EParamName {eTheta=0, eRho=1, eGrowth=2, eMeanTractLen=3, eNone};//important: keep eNone as last!
typedef vector<int> TParamNames;

/*---------------------------------------------------------------------------*/

enum E2LocusCat { e00=0, e01, e10, e11, e0x, e1x, ex0, ex1 };
bool IsMissing(E2LocusCat c) {
    return c==e0x  ||  c==e1x  ||  c==ex0  ||  c==ex1;
}
ostream& operator<< (ostream &o, const unsigned char v) {
    o << (int) v; return o;
}
istream& operator>> (istream &i, unsigned char& v) {
    int a; i >> a; v = a; return i;
}
class T2LocusConfig {
public:
    T2LocusConfig(void) { clear(); }
    void clear(void) { memset(&m_Cfg, 0, sizeof(m_Cfg)); }
    const unsigned char operator[] (unsigned int i) const { return m_Cfg[i]; }
    unsigned char& operator[] (unsigned int i) { return m_Cfg[i]; }
    unsigned int N(void) const {
        return (*this)[e00] + (*this)[e01] + (*this)[e10] + (*this)[e11] +
            (*this)[e0x] + (*this)[e1x] + (*this)[ex0] + (*this)[ex1];
    }
    unsigned int Nmissing(void) const {
        return (*this)[e0x] + (*this)[e1x] + (*this)[ex0] + (*this)[ex1];
    }
    unsigned int Lmissing(void) const { return (*this)[ex0] + (*this)[ex1]; }
    unsigned int Rmissing(void) const { return (*this)[e0x] + (*this)[e1x]; }
    bool BothFixed(void) const {
        int n = N();
        return ((*this)[e00] + (*this)[e0x] + (*this)[ex0] == n)  ||
            ((*this)[e01] + (*this)[e0x] + (*this)[ex1] == n)  ||
            ((*this)[e10] + (*this)[e1x] + (*this)[ex0] == n)  ||
            ((*this)[e11] + (*this)[e1x] + (*this)[ex1] == n);
    }
    bool BothSeg(void) const {
        return ((*this)[e00] + (*this)[e01] + (*this)[e0x] > 0)  &&
            ((*this)[e00] + (*this)[e10] + (*this)[ex0] > 0)  &&
            ((*this)[e11] + (*this)[e10] + (*this)[e1x] > 0)  &&
            ((*this)[e11] + (*this)[e01] + (*this)[ex1] > 0);
    }

    double LogNumOrderedCfgs(void) const {
        return gsl_sf_lnfact(N()) - 
            gsl_sf_lnfact((*this)[e00]) -
            gsl_sf_lnfact((*this)[e01]) -
            gsl_sf_lnfact((*this)[e10]) -
            gsl_sf_lnfact((*this)[e11]) -
            gsl_sf_lnfact((*this)[e0x]) - 
            gsl_sf_lnfact((*this)[e1x]) -
            gsl_sf_lnfact((*this)[ex0]) -
            gsl_sf_lnfact((*this)[ex1]);
    }
    bool operator< (const T2LocusConfig &c) const {
        return memcmp(this, &c, sizeof(T2LocusConfig)) < 0;
    }
    bool operator== (const T2LocusConfig &c) const {
        return memcmp(this, &c, sizeof(T2LocusConfig)) == 0;
    }
    struct hash { size_t operator()(const T2LocusConfig& c) const {
        // Fowler / Noll / Vo (aka FNV) hash
        // (type FNV-1a; code from g++ 'functional' header 4.0.2)
        const char* first = reinterpret_cast<const char*>(c.m_Cfg);
        size_t length = sizeof(c.m_Cfg);
        size_t result = 2166136261UL;
        for (; length > 0; --length)
          {
            result ^= (std::size_t)*first++;
            result *= 16777619UL;
          }
        return result;
    }};
    friend struct hash;

private:
    unsigned char m_Cfg[8];
};
ostream& operator<< (ostream &o, const T2LocusConfig& cfg) {
    for(int i = 0;  i < 8;  ++i) {
        if (i != 0) {
            o << ".";
        }
        o << (int) cfg[i];
    }
    return o;
}

/*---------------------------------------------------------------------------*/

struct SBase {
    const CChr *m_Chr;
    ENuc m_Nt;
    unsigned short m_Qual;
    double GetErrProb(void) const {
        return pow(10,-(double)m_Qual/10.0);
    }
    bool operator< (const SBase &x) const {
        return m_Chr < x.m_Chr;
    }
    SBase(const CChr *c, ENuc n, unsigned short q) :
        m_Chr(c), m_Nt(n), m_Qual(q) {}

    class CQualCmp {
    public: bool operator() (const SBase &a, const SBase &b) {
        return a.m_Qual < b.m_Qual;
    }};
};
typedef vector< SBase > TAlnCol;
typedef vector< const SBase* > TAlnColSubset;

/*---------------------------------------------------------------------------*/
// n choose k items
class CCombination : public vector<bool> {
public:
    CCombination(unsigned int n, unsigned int k) :
        vector<bool>(n), m_NumOnBits(k), m_RolledOver(false), m_Packed(true),
        m_NextPotential10(k) {
        fill(begin(), begin()+k, true);
        fill(begin()+k, end(), false);
    }
    CCombination& operator++ (void);
    bool good(void) const { return !m_RolledOver; }
    bool IsPacked(void) const { return m_Packed; }
    void AdvanceBeforeNextBiggest(void);
    friend ostream& operator<< (ostream &o, const CCombination &c) {
        for (const_iterator it = c.begin();  it != c.end();  ++it) {
            o << (*it ? 1:0);
        }
        return o;
    }
private:
    unsigned int m_NumOnBits;
    bool m_RolledOver;
    bool m_Packed;
    unsigned short m_NextPotential10;
};

/*---------------------------------------------------------------------------*/

class CAssm_CI {
public:
    static int    sm_QualCutoff;
    static double sm_ErrCutoff;//calculated from sm_QualCutoff
    static unsigned int sm_MaxBadBases;
    enum ESNPStatus {
        eUnknown,
        eFixed,
        eSNP,
        eSingletonSNP
    };
    enum EWhichSites {
        eAll,
        eSeg,
        eSegNoSingle
    };
    class CCol {
    public:
        CCol(void) : m_IsSNP(eUnknown) {}
        double GetPrObservedFS(const CFreqSpectrum &fs, bool trustObs) const;
        const TAlnCol& GetCol(void) const { return m_Col; }
        TAlnCol& SetCol(void) { return m_Col; }
        ESNPStatus IsSNP(void) const { return m_IsSNP; }
        ESNPStatus& SetIsSNP(void) { return m_IsSNP; }
        unsigned int GetPos(void) const { return m_Pos; }
        void SetPos(unsigned int p) { m_Pos = p; }

        friend ostream& operator<< (ostream &o, const CCol &c) {
            for (TAlnCol::const_iterator i = c.GetCol().begin();
                 i != c.GetCol().end();  ++i) {
                if (i != c.GetCol().begin()) {
                    o << "\t";
                }
                o << Enum2Ch(i->m_Nt) << '(' << i->m_Qual << ')';
                //o << "/" << i->m_Chr->GetReads().front().GetAcc();
            }
            return o;
        }

        void WriteFSDist(ostream&) const;
        void ReadFSDist(istream&);

    private: //this block for calculating frequency spectrum probabilities
        typedef vector<double> TProbDist;
        void x_CalcErrorDist(const TAlnColSubset &chrs, TProbDist &errorPrbs)
            const;
        void x_CalcObservedFSCondDist(void);

    private:
        unsigned int m_Pos; //really just for debugging so can report pos.
        ESNPStatus m_IsSNP;
        mutable TAlnCol m_Col;
        mutable TProbDist m_prObservedCond;
    };

public:
    CAssm_CI(const CAssembly &assm, unsigned int start, unsigned int stop,
             EWhichSites which) :
        m_Assm(assm), m_Start(start), m_Stop(stop), m_WhichSites(which),
        m_Pos(m_Start), m_ChrRangeStart(0), m_ChrRangeStop(0)
    {
        m_ThisIsCopy = false;
        m_Cols = new CCol[m_Stop-m_Start+1];
        m_ReadIdx = new TReadIdx;
        for (CAssembly::const_iterator chrI = assm.begin();
             chrI != assm.end();  ++chrI){
            for (CChr::TReads::const_iterator readI = (*chrI)->GetReads().
                     begin();  readI != (*chrI)->GetReads().end();  ++readI) {
                m_ReadIdx->push_back(&*readI);
            }
        }
        sort(m_ReadIdx->begin(), m_ReadIdx->end(), CRead::CPtrCmp());
        m_RdBegin = m_RdEnd = m_ReadIdx->begin();
        --m_Pos; ++(*this); //trick to advance to correct first position
        //x_AdvanceRdRange();
    }

    // Make a quick copy that we do NOT own -- left up to user to make sure
    // that original is not deleted prematurely.
    CAssm_CI(const CAssm_CI& it) : m_Assm(it.m_Assm) {
        memcpy(this, &it, sizeof(CAssm_CI));
        m_ThisIsCopy = true;
    }
    ~CAssm_CI(void) {
        if (!m_ThisIsCopy) { delete[] m_Cols; delete m_ReadIdx; }
    }

    bool good(void) const { return m_Pos <= m_Stop; }

    //    void Reset(void) { m_CachePos = m_SNPs->begin(); }
    CAssm_CI& operator++(void) {
        ++m_Pos; x_AdvanceRdRange();
        if (m_WhichSites == eSeg) {
            while (good()  &&  !IsSNPPos()) { ++m_Pos; x_AdvanceRdRange(); }
        } else if (m_WhichSites == eSegNoSingle) {
            while (good() &&  !((**this).IsSNP() == eSNP)) {
                ++m_Pos; x_AdvanceRdRange();
            }
        }
        return *this;
    }
    CAssm_CI& operator+=(unsigned int n) {
        if (m_WhichSites == eAll) {
            m_Pos += n; x_AdvanceRdRange();
        } else {
            for (unsigned int i = 0;  i < n;  ++i) { ++(*this); }
        }
        return *this;
    }
    const CAssm_CI operator+(unsigned int n) const {
        CAssm_CI it(*this);
        return it += n;
    }
    void AdvancePast(unsigned int pos) {
        assert(m_Pos < pos  &&  pos <= m_Stop);
        m_Pos = pos - 1;
        ++(*this);
    }

    const CCol& operator* (void) const;
    const CCol* operator-> (void) const { return &(**this); };

    unsigned int GetPos(void) const { return m_Pos; }
    //NOTE: the "depth" below includes gaps and Ns
    unsigned int GetDepth(void) const { return m_RdEnd - m_RdBegin; }
    bool IsSNPPos(void) const;
    /*
    bool ChrOverlapsWith(const CAssm_CI &i) const {
        return x_GetChrRangeStart() <= i.GetPos()  &&
            i.GetPos() <= x_GetChrRangeStop();
    }
    */
    bool RdOverlapsWith(const CAssm_CI &i) const {
        //note this is not quite right, since we sort by start, not by stop
        return (*m_RdBegin)->GetStart() <= i.GetPos()  &&
            i.GetPos () <= (*(m_RdEnd-1))->GetStop();
    }

    void LoadFSDist(istream &in) {
        string buff;
        while (getline(in, buff).good()) {
            istringstream iss(buff);
            unsigned int pos;
            iss >> pos;
            if (pos >= m_Start  &&  pos <= m_Stop) {
                m_Cols[pos-m_Start].ReadFSDist(iss);
            }
        }
    }

    static void SetQualThresholds(unsigned int qCutoff, unsigned int maxBad) {
        sm_QualCutoff = qCutoff;
        sm_ErrCutoff = //1e-6 for numerical gt
            pow(10, -(double)sm_QualCutoff/10) + 1e-6;
        sm_MaxBadBases = maxBad;
    }

protected:
    void x_AdvanceRdRange(void);
    /*
    unsigned int x_GetChrRangeStart(void) const { 
        return m_ChrRangeStart; }
    unsigned int x_GetChrRangeStop(void) const {
        return m_ChrRangeStop; }
    */
protected:
    typedef vector<const CRead*> TReadIdx;

    bool m_ThisIsCopy;
    const CAssembly &m_Assm;
    unsigned int m_Start, m_Stop;
    EWhichSites m_WhichSites;
    mutable CCol *m_Cols;

    unsigned int m_Pos;
    TReadIdx *m_ReadIdx;
    TReadIdx::const_iterator m_RdBegin, m_RdEnd;
    unsigned int m_ChrRangeStart, m_ChrRangeStop;
};
//Set some heuristics here -- how many bases are we willing to
//tolerate at a given site (or pair of sites) with quality worse than
//the specified threshold?
int CAssm_CI::sm_QualCutoff = 20;
double CAssm_CI::sm_ErrCutoff = 
    pow(10, -(double)CAssm_CI::sm_QualCutoff/10) + 1e-6;
unsigned int CAssm_CI::sm_MaxBadBases = 5;

class Csnp_CI : public CAssm_CI {
public:
    Csnp_CI(const CAssm_CI& it) : CAssm_CI(it) {
        if (!IsSNPPos()) { ++(*this); }
    }
    Csnp_CI(const Csnp_CI& it) : CAssm_CI(it) {}
    Csnp_CI& operator++(void) {
        if (good()) {
            do { CAssm_CI::operator++(); } while (good()  &&  !IsSNPPos());
        }
        return *this;
    }
    const Csnp_CI operator+(unsigned int n) const {
        Csnp_CI it(*this);
        for (; n > 0;  --n, ++it) {}
        return it;
    }
};

class CAssmOverlap_CI : public CAssm_CI {
public:
    CAssmOverlap_CI(const CAssm_CI& assmI) : CAssm_CI(assmI) {
        TReadIdx readIdx;
        readIdx.reserve(2*(m_RdEnd-m_RdBegin));
        for (TReadIdx::const_iterator it = m_RdBegin;  it != m_RdEnd;  ++it) {
            const CChr::TReads &reads = (*it)->GetChr()->GetReads();
            for (CChr::TReads::const_iterator readI = reads.begin();
                 readI != reads.end();  ++readI) {
                readIdx.push_back(&*readI);
            }
        }
        sort(readIdx.begin(), readIdx.end(), CRead::CPtrCmp());
        if (!readIdx.empty()) {
            m_Intervals.push_back(*readIdx.front());
            for (TReadIdx::const_iterator it = readIdx.begin()+1;
                 it != readIdx.end();  ++it) {
                if (m_Intervals.back().OverlapsOrAdjacent(**it)) {
                    m_Intervals.back().Union(**it);
                } else {
                    m_Intervals.push_back(**it);
                }
            }
        }

        for (m_CurrInterval = m_Intervals.begin();
             m_CurrInterval != m_Intervals.end()  &&
                 !m_CurrInterval->Overlaps(GetPos());
             ++m_CurrInterval) {}
    }
    CAssmOverlap_CI& operator++(void) {
        CAssm_CI::operator++();
        do {
            while (m_CurrInterval != m_Intervals.end()  &&
                   GetPos() > m_CurrInterval->m_Stop) {
                ++m_CurrInterval;
            }
            if (m_CurrInterval != m_Intervals.end()  &&
                GetPos() < m_CurrInterval->m_Start) {
                AdvancePast(m_CurrInterval->m_Start);
            }
        } while (m_CurrInterval != m_Intervals.end()  &&
                 !m_CurrInterval->Overlaps(GetPos()));
        return *this;
    }
    bool good(void) const {
        return CAssm_CI::good() &&  m_CurrInterval != m_Intervals.end(); }

private:
    struct SInterval {
        unsigned int m_Start, m_Stop;
        SInterval(const CRead &r):m_Start(r.GetStart()), m_Stop(r.GetStop()) {}
        bool Overlaps(unsigned int p) const {
            return m_Start <= p  &&  m_Stop >= p;
        }
        bool OverlapsOrAdjacent(const CRead &r) const {
            //note don't want use subtraction because unsigned
            return m_Start <= r.GetStop()+1  &&  m_Stop+1 >= r.GetStart();
        }
        void Union(const CRead &r) {
            m_Start = min(m_Start, r.GetStart());
            m_Stop  = max(m_Stop,  r.GetStop());
        }
    };
    typedef vector<SInterval> TIntervals;
    TIntervals m_Intervals;
    TIntervals::const_iterator m_CurrInterval;
};

/*---------------------------------------------------------------------------*/
class C2LocusProbs {
    typedef vector<double> TProbs;    //all rho values for one configuration

    //multiple pairtypes
    class TCfgTable : 
#ifdef SYSTEM_MISSING_TR1
        public map<T2LocusConfig, TProbs> {
        typedef map<T2LocusConfig, TProbs> TParent;
#else
        public tr1::unordered_map<T2LocusConfig, TProbs, T2LocusConfig::hash> {
        typedef tr1::unordered_map<T2LocusConfig, TProbs,
                                   T2LocusConfig::hash> TParent;
#endif
    public:
        const TProbs& find(const T2LocusConfig &cfg) const {
            const_iterator it = TParent::find(cfg);
            if (it == end()) {
                ostringstream oss; oss << cfg;
                throw runtime_error("missed cache: " + oss.str());
            } else {
                return it->second;
            }
        }
        const TProbs* Get(const T2LocusConfig &cfg) const {
            try { return &((*this)[cfg]); } catch (...) { return NULL; }
        }
        const TProbs& operator[] (const T2LocusConfig &cfg) const {
            return find(x_Canon(cfg));
        }
        TProbs& Set(const T2LocusConfig &cfg) {
            return TParent::operator[](x_Canon(cfg));
        }
        /*
        friend ostream& operator<< (ostream &o, const TCfgTable &t) {
            for (const_iterator cfgI = t.begin();  cfgI != t.end();  ++cfgI) {
                o << cfgI->first;
                for (TProbs::const_iterator rhoI = cfgI->second.begin();
                     rhoI != cfgI->second.end();  ++rhoI) {
                    o << "\t" << *rhoI;
                }
                o << endl;
            }
            return o;
        }
        */
        void ReadBinary(istream &i) {
            short int n;
            T2LocusConfig cfg;
            while (i.read(reinterpret_cast<char*>(&cfg), sizeof(cfg)).good()) {
                i.read(reinterpret_cast<char*>(&n), sizeof(n));
                TProbs &probs = Set(cfg);
                //NOTE: I make the assumption that memory for the vector
                //is contiguous starting at front (*not* specified in the
                //standard)
                probs.resize(n);
                i.read(reinterpret_cast<char*>(&probs.front()),
                       sizeof(TProbs::value_type)*n);
            }
        }
        void WriteBinary(ostream &o) const {
            for (const_iterator cfgI = begin();  cfgI != end();  ++cfgI) {
                o.write(reinterpret_cast<const char*>(&cfgI->first),
                        sizeof(cfgI->first));
                short int n = cfgI->second.size();
                o.write(reinterpret_cast<const char*>(&n), sizeof(n));
                o.write(reinterpret_cast<const char*>(&cfgI->second.front()),
                        sizeof(TProbs::value_type)*cfgI->second.size());
            }
        }
    private:
        const T2LocusConfig& x_Canon(const T2LocusConfig &cfg) const {
            return (cfg[ex0] + cfg[ex1] + cfg[e01] <
                    cfg[e0x] + cfg[e1x] + cfg[e10]) ?
                x_Swap(cfg) : cfg;
        }
        const T2LocusConfig& x_Swap(const T2LocusConfig &cfg) const {
            m_Tmp[e00] = cfg[e00]; m_Tmp[e01] = cfg[e10];
            m_Tmp[e10] = cfg[e01]; m_Tmp[e11] = cfg[e11];
            m_Tmp[e0x] = cfg[ex0]; m_Tmp[ex0] = cfg[e0x];
            m_Tmp[e1x] = cfg[ex1]; m_Tmp[ex1] = cfg[e1x];
            return m_Tmp;
        }
    private:
        mutable T2LocusConfig m_Tmp;
    };
    /*
    class TCfgTable : public CCache<TProbs, T2LocusConfig> {
        typedef CCache<TProbs, T2LocusConfig> TParent;
    public: //automatically switch to canonical cfg
        const TProbs& operator[] (const T2LocusConfig &cfg) const {
            return TParent::operator[]((cfg[e01] < cfg[e10]) ?
                                       x_Swap(cfg) : cfg);
        }
        TProbs& Set(const T2LocusConfig &cfg) {
            return TParent::Set[]((cfg[e01] < cfg[e10]) ?
                                  x_Swap(cfg) : cfg);
        }
    private:
        const T2LocusConfig& x_Swap(const T2LocusConfig &cfg) {
            m_Tmp[e00] = cfg[e00]; m_Tmp[e01] = cfg[e10];
            m_Tmp[e10] = cfg[e01]; m_Tmp[e11] = cfg[e11];
            return m_Tmp;
        }
    private:
        mutable T2LocusConfig m_Tmp;
    };
    */
    struct TDepthProbs {
        TCfgTable m_CfgProbs;
        TProbs m_NormProbs;
        //below used for adjusting probs to different thetas
        TProbs m_SumSegSeg; // sum over all dual segregating cfgs
        TProbs m_SumSegFix; // sum over all singly-segregating cfgs
        double m_A1; //\sum_{i=1}^{n-1} 1/i
    };
public:
    enum ELookupTableFmt {
        eMcVean,
        eHudson,
        eAuto
    };
    typedef CAssm_CI::EWhichSites EWhichSites;
    typedef vector<double> TRhoValues;

public:
    C2LocusProbs(istream &, EWhichSites whichSites, ELookupTableFmt = eAuto);
    void LoadMcVean(istream &);
    void LoadHudson(istream &);

    void LoadMissingNormConsts(istream &);

    void KickNormConst(unsigned int n) const;
    void KickCacheMissing(unsigned int n, const CParams &) const;
    void SlightlyLessTrivialMax(double trueRho, unsigned int n) const;
    void TrivialMax(unsigned int trueRhoI, unsigned int n) const;
    void TestMiss(void) const;
    void DumpPr(unsigned int n, unsigned int L, unsigned int R, bool fold,
                unsigned int qual) const;
    void DumpPrFixed(unsigned int n) const;
    double x_CalcScore(const CParams &p, const T2LocusConfig &cfg,
                       bool fold) const;
    double x_CalcFI(const CParams &p, unsigned int n, unsigned int L,
                    unsigned int R, bool fold) const;
    void DumpFisherInformation(unsigned int n, bool fold) const;
    void DumpFisherInformation(CParams p, unsigned int n,
                               unsigned int L, unsigned int R) const;
    void DumpFIp_s(CParams p, unsigned int n) const;
    void DumpFisherInformationAFS(unsigned int n, bool fold) const;
    void DumpFisherInformationSingletons(unsigned int n, bool fold) const;
    void DumpFisherInformationNonSingletons(unsigned int n, bool fold) const;
    void DumpSegProbs(unsigned int n) const;

    //store/load currently cached missing probabilities
    void WriteMissingPrTable(ostream &o) const {m_MissingProbs.WriteBinary(o);}
    void ReadMissingPrTable(istream &i) {m_MissingProbs.ReadBinary(i);}

    const TRhoValues& GetTabulatedRhoValues(void) const { return m_RhoValues; }
    double GetTableTheta(void) const { return m_TableTheta; }
    unsigned int GetMaxDepth(void) const { return m_MaxDepth; }

    double GetFoldedPr(const CParams &p, const T2LocusConfig& cfg) const;
    double GetPrMissing(const CParams &p, const T2LocusConfig& cfg) const;
    double GetPrComplete(const CParams &p, const T2LocusConfig& cfg) const {
        assert(cfg.Nmissing() == 0);
        const TProbs &prbs = x_GetDepthProbs(cfg.N()).m_CfgProbs[cfg];
        unsigned int rhoPos = x_GetFloorRhoPos(p.Get(eRho));
        bool beyondEnd = false;
        if (rhoPos == m_RhoValues.size()-1) {//assume const beyond end
            beyondEnd = true;
            --rhoPos;
        }

        double probL = prbs[rhoPos];
        double probH = prbs[rhoPos+1];
        if (m_WhichSites != CAssm_CI::eAll) {
            const TProbs &norms = x_GetDepthProbs(cfg.N()).m_NormProbs;
            probL -= norms[rhoPos];
            probH -= norms[rhoPos+1];
        } else if (p.Get(eTheta) != m_TableTheta) {
            //approximate adjustment for desired theta
            const TDepthProbs &dp = x_GetDepthProbs(cfg.N());
            double newSegProb = 1 - exp(-p.Get(eTheta) * dp.m_A1);
            double oldSegProb = 1 - exp(-m_TableTheta * dp.m_A1);
            if (cfg.BothFixed()) {
                double ssAdj = log(newSegProb * newSegProb /
                                   (oldSegProb * oldSegProb));
                double sfAdj = log(newSegProb * (1-newSegProb) /
                                   (oldSegProb * (1-oldSegProb)));
                probL = SubLogged(log(1),
                                  SumLogged(dp.m_SumSegSeg[rhoPos] + ssAdj,
                                            dp.m_SumSegFix[rhoPos] + sfAdj)) -
                    log(4);
                probH = SubLogged(log(1),
                                  SumLogged(dp.m_SumSegSeg[rhoPos+1] + ssAdj,
                                            dp.m_SumSegFix[rhoPos+1] + sfAdj))-
                    log(4);
            } else if (cfg.BothSeg()) {
                double adj = log(newSegProb * newSegProb /
                                 (oldSegProb * oldSegProb));
                probL += adj;
                probH += adj;
            } else {
                double adj = log(newSegProb * (1-newSegProb) /
                                 (oldSegProb * (1-oldSegProb)));
                probL += adj;
                probH += adj;
            }
        }

        if (beyondEnd) {
            return probH; //assume const beyond end
        }

        //not sure this is correct, but Hudson does interpolation in logspace whenever possible, so we'll do the same for now.
        if (m_RhoValues[rhoPos] == 0  &&  isinf(prbs[rhoPos])) {
            return log(x_Interpolate(m_RhoValues[rhoPos],m_RhoValues[rhoPos+1],
                                     exp(probL), exp(probH),
                                     p.Get(eRho)));
        } else if (m_RhoValues[rhoPos] == 0) {
            return x_Interpolate(m_RhoValues[rhoPos], m_RhoValues[rhoPos+1],
                                 probL, probH,
                                 p.Get(eRho));
        } else {
            return x_Interpolate(log(m_RhoValues[rhoPos]),
                                 log(m_RhoValues[rhoPos+1]),
                                 probL, probH,
                                 log(p.Get(eRho)));
        }
    }

private:
    void x_ChangeTheta(double newTheta, unsigned int n);
    void x_CalcPrMissing(CParams tmpP, const T2LocusConfig &obs, TProbs &probs)
        const;
    void x_CacheNormPrMissing(const CParams &p, unsigned int L, unsigned int R,
                              unsigned int n) const;
    void x_RemoveSinglesFromNorms(TProbs &norm, const CParams &p,
                                  unsigned int L, unsigned int R,
                                  unsigned int n) const;
    double x_PrObsCondCmpNorm(unsigned int L, unsigned int R,
                              const T2LocusConfig &cmp) const;
    double x_PrObsCondCmp(const T2LocusConfig &obs,
                          const T2LocusConfig &cmp) const;

    enum ERestrict {
        eNoLR = 1
    };
    unsigned int x_Cfg2Pairtype(const T2LocusConfig &cfg,
                                unsigned int restrict = 0) const;
    unsigned int x_Cfg2Pairtype(unsigned int a, unsigned int b,
                                unsigned int c, unsigned int d) const {
        T2LocusConfig cfg;
        cfg[e00]=a; cfg[e01]=b; cfg[e10]=c; cfg[e11]=d;
        return x_Cfg2Pairtype(cfg);
    }
    const T2LocusConfig& x_Pairtype2Cfg(unsigned int pt, unsigned int n) const;
    unsigned int x_MaxPairtype(unsigned int n) const;
    unsigned int x_GetFloorRhoPos(double) const;
    const TDepthProbs& x_Downsample(unsigned int n);
    void x_CalcNormConst(unsigned int n);

    const TDepthProbs& x_GetDepthProbs(unsigned int n) const {
        try {
            return m_2LocusProbs[n];
        } catch (...) {//missed cache
            return const_cast<C2LocusProbs*>(this)->x_Downsample(n);
        }
    }

    double x_Interpolate(double x1, double x2, double y1, double y2, double x)
        const { return ((x - x1)/(x2-x1)) * (y2-y1) + y1; }
private:
    EWhichSites m_WhichSites;
    unsigned int m_MaxDepth;
    TRhoValues m_RhoValues;
    double m_TableTheta; // theta used to generate likelihood table
    
    //cache probs & normalization factors (to get prob conditional on both sites being segregating) by depth
    mutable CCache< TDepthProbs > m_2LocusProbs; //cache by depth
    //cache pairtype -> cfg mapping --> mainly for debug(?)
    mutable vector< CCache< T2LocusConfig > > m_Pairtype2Cfg; 

    //cache probs for missing data
    mutable TCfgTable m_MissingProbs;

    //cache missing data norm. const
    struct SMissing {
        unsigned int m_L, m_R, m_N;
        SMissing(void) {}
        SMissing(unsigned int l, unsigned int r, unsigned int n) :
            m_L(l), m_R(r), m_N(n) { if (l < r) { swap(m_L, m_R); } }
        bool operator< (const SMissing &a) const {
            return memcmp(this, &a, sizeof(SMissing)) < 0;
        }
        friend ostream& operator<< (ostream &o, const SMissing &s) {
            o << s.m_N << ":" << s.m_L << "," << s.m_R;
            return o;
        }
    };
    mutable CCache< TProbs, SMissing> m_MissingNormConsts;

    // BELOW is JUNK waiting to be deleted
    //cache # ways to lose (for missing data norm. const)
    struct SWays2Lose {
        SWays2Lose(unsigned int l, unsigned int r, const T2LocusConfig& cfg) :
            m_L(l), m_R(r), m_Cfg(cfg) {}
        friend ostream& operator<< (ostream &o, const SWays2Lose &s) {
            o << s.m_L << "," << s.m_R << " --> " << s.m_Cfg;
            return o;
        }
        bool operator< (const SWays2Lose &a) const {
            return m_L < a.m_L  ||
                (m_L == a.m_L  && (m_R < a.m_R  ||
                                   (m_R == a.m_R  &&  m_Cfg < a.m_Cfg)));
        }
        unsigned int m_L;
        unsigned int m_R;
        T2LocusConfig m_Cfg; //*complete* configuration
    };
    mutable CCache< double, SWays2Lose > m_Ways2Lose;
};

/*---------------------------------------------------------------------------*/

class CPIIMmodel : public CLikelihoodModel {
public:
    typedef vector< pair<T2LocusConfig, double> > TProbCfgs;
    struct SSNPPair {
        unsigned int  m_Dist;//distance separating the two SNPs
        TProbCfgs m_ProbCfgs;

        void WriteBinary(CFileBuffer &out) const {
            if (out.write(reinterpret_cast<const char*>(&m_Dist),
                          sizeof(m_Dist)) == 0) {
                throw runtime_error("WriteBinary writing failed");
            }
            short unsigned int numCfgs = m_ProbCfgs.size();
            out.write(reinterpret_cast<const char*>(&numCfgs),sizeof(numCfgs));
            //NOTE: I make the assumption that memory for the vector
            //is contiguous starting at front (*not* specified in the
            //standard)
            out.write(reinterpret_cast<const char*>(&m_ProbCfgs.front()),
                      sizeof(TProbCfgs::value_type)* m_ProbCfgs.size());
            /*
            for (TProbCfgs::const_iterator it = m_ProbCfgs.begin();
                 it != m_ProbCfgs.end();  ++it) {
                fwrite(reinterpret_cast<const char*>(&*it), sizeof(*it), 1,
                       out);
            }
            */
        }
        void ReadBinary(CFileBuffer &in) {
            if (in.read(reinterpret_cast<char*>(&m_Dist), sizeof(m_Dist))
                == 0) {
                throw runtime_error("no bytes read by ReadBinary");
            }
            short unsigned int numCfgs;
            in.read(reinterpret_cast<char*>(&numCfgs), sizeof(numCfgs));
            //NOTE: I make the assumption that memory for the vector
            //is contiguous starting at front (*not* specified in the
            //standard)
            m_ProbCfgs.resize(numCfgs);
            in.read(reinterpret_cast<char*>(&m_ProbCfgs.front()),
                    sizeof(TProbCfgs::value_type)*numCfgs);
            /*
            m_ProbCfgs.clear();
            TProbCfgs::value_type pr;
            for (int j = 0;  j < numCfgs;  ++j) {
                fread(reinterpret_cast<char*>(&pr), sizeof(pr), 1, in);
                m_ProbCfgs.insert(pr);
            }
            */
        }
    };
#ifdef DISKCACHE
    typedef CDiskList< SSNPPair, (unsigned int)1e8 > TSNPPairs;
#else
    typedef vector< SSNPPair > TSNPPairs;
#endif
    enum ERecombModel {
        eGeneConv,
        eCrossover
    };

public:
    CPIIMmodel(istream &inProbs, ERecombModel recombModel,
               C2LocusProbs::EWhichSites whichSites, int minPairDist) :
        m_2LocusProbs(inProbs, whichSites), m_RecombModel(recombModel),
        m_WhichSites(whichSites), m_MinPairDist(minPairDist),
        m_TrustObservations(false) {}

    void LoadAssembly(istream &in, int start = -1, int stop = -1) {
#ifdef MONITOR_LOAD
        cerr << "Loading assembly..\r";
#endif
        m_Assm.Load(in);
        if (start == -1  ||  start < (int) m_Assm.GetStart()) {
            start = m_Assm.GetStart();
        }
        if (stop == -1  ||  stop > (int) m_Assm.GetStop()) {
            stop = m_Assm.GetStop();
        }
        m_AssmIt.reset(new CAssm_CI(m_Assm, start, stop, m_WhichSites));
        m_SNPpairs.reset();
#ifdef MONITOR_LOAD
        cerr << "Loaded assembly     " << endl;
#endif
    }
    void SaveAssembly(ostream &o) const {
        m_Assm.Save(o);
    }
    void LoadNormConsts(istream &in) {
        m_2LocusProbs.LoadMissingNormConsts(in);
    }
    void LoadFSDist(istream &in) {
        if (!m_AssmIt.get()) {
            throw runtime_error("Attempted to load FS dist but no assembly loaded!");
        }
        m_AssmIt->LoadFSDist(in);
    }
    void LoadCachedMissing(istream &in) {
        m_2LocusProbs.ReadMissingPrTable(in);
    }

    //DEBUG REMOVE ME BELOW
    void LoadMaxhap(istream &in);
    void DumpPairPrs(CParams p) const;
    void DepthHist(void) const;
    void PrintObservedFS(void) const;
    void PrintPos(unsigned int p) const;
    void PrintCoverage(void) const;
    void AverageP_s(void) const;
    unsigned int x_getchrlength(void) {
        return m_Assm[0].GetStop() - m_Assm[0].GetStart() +1;
    }

    //trust observations (and ignore q scores)?
    void SetTrustObs(bool trust) { m_TrustObservations = trust; }

    //use frequency spectrum? (instead of pairs()
    void SetFreqSpec(const string& s) { m_FreqSpec = s; }

    //total likelihood surface
    typedef vector< pair<CParams, double> > TLikSurface;
    virtual void CalcLogLikSurface(const CParams&, TLikSurface&) const;

    //total likelihood
    virtual double CalcLogLik(const CParams&) const;
    //per-pair likelihood
    double CalcPairLogLik(const CParams &p, const T2LocusConfig &c) const {
        return m_2LocusProbs.GetFoldedPr(p, c);
        //return m_2LocusProbs.GetPrComplete(p, c, false);
        //return m_2LocusProbs.GetPrMissing(p, c); //DEBUG
    }

    const CAssm_CI& GetAssmIterator(void) const {
        if (!m_AssmIt.get()) {
            throw runtime_error("No assembly loaded!");
        }
        return *m_AssmIt;
    }
    const TSNPPairs& GetSNPPairs(void) const {
        if (!m_SNPpairs.get()) {
            m_SNPpairs.reset(new TSNPPairs);
            const_cast<CPIIMmodel*>(this)->x_CacheSNPPairs();
        }
        return *m_SNPpairs;
    }
    const C2LocusProbs& Get2LocusProbs(void) const { return m_2LocusProbs; }

protected:
    void x_CacheSNPPairs(void);
public://debug -- called from DumpPr
    static void x_CalcObsProbCfgs(const CAssm_CI&, const CAssm_CI&,
                                  SSNPPair &snpPair);

private:
    CAssembly m_Assm;
    C2LocusProbs m_2LocusProbs;
    unique_ptr<CAssm_CI> m_AssmIt;
    mutable unique_ptr<TSNPPairs> m_SNPpairs;
    string m_FreqSpec;
    ERecombModel m_RecombModel;
    CAssm_CI::EWhichSites m_WhichSites;
    int m_MinPairDist;
    bool m_TrustObservations; //if true, ignore q scores

public:
    //when the probability of a configuration drops below this
    //fraction of the maximal probability, we will ignore it
    static double sm_SmallEnough;
    //when picking pairs of sites, increment by this much in outer
    //loop (really just used for thinning data)
    static unsigned int sm_SiteIncrement;
};
double CPIIMmodel::sm_SmallEnough;
unsigned int CPIIMmodel::sm_SiteIncrement;

/*---------------------------------------------------------------------------*/
// PRE : iterator properly initialized
// POST: a "column" representation of the current position (also this
// is now cached)
const CAssm_CI::CCol& CAssm_CI::operator*(void) const {
    CCol& col = m_Cols[m_Pos-m_Start];

    if (col.GetCol().empty()) {
        unsigned int numBelowThreshold = 0;

        col.SetPos(m_Pos);
        col.SetIsSNP() = eFixed;
        ENuc firstAllele = eGap; //just to silence compiler warning
        for (TReadIdx::const_iterator readI = m_RdBegin;  readI != m_RdEnd;
             ++readI){
            if ((*readI)->Overlaps(GetPos())  &&
                (*readI)->GetBase(GetPos()) != eGap  &&
                (*readI)->GetBase(GetPos()) != eN) {
                if (col.GetCol().empty()) {
                    firstAllele = (*readI)->GetBase(GetPos());
                } else if ((*readI)->GetBase(GetPos()) != firstAllele)  {
                    col.SetIsSNP() = eSNP;
                }
                if ((*readI)->GetQual(GetPos()) < sm_QualCutoff) {
                    ++numBelowThreshold;
                }
                col.SetCol().push_back(TAlnCol::value_type
                                       ((*readI)->GetChr(),
                                        (*readI)->GetBase(GetPos()),
                                        (*readI)->GetQual(GetPos())));
            }
        }

        if (numBelowThreshold > sm_MaxBadBases) {
            //eliminate the worst bases until we only have
            //sm_MaxBadBases worse than the threshold
            sort(col.SetCol().begin(), col.SetCol().end(), SBase::CQualCmp());
            TAlnCol::iterator firstGoodIt;
            for (firstGoodIt = col.SetCol().begin();
                 firstGoodIt != col.SetCol().end()  &&
                     firstGoodIt->m_Qual < sm_QualCutoff;  ++firstGoodIt) {}
            col.SetCol().erase(col.SetCol().begin(),
                               firstGoodIt-sm_MaxBadBases);
        }

        //minorize tri & quad allelic sites
        if (col.IsSNP() == eSNP) {
            unsigned int numAlleles[4] = {0,0,0,0};
            for (TAlnCol::iterator it = col.SetCol().begin();
                 it != col.SetCol().end();  ++it) {
                ++numAlleles[it->m_Nt];
            }
            int major = eA;
            for (int i = 0;  i < 4;  ++i) {
                if (numAlleles[i] > numAlleles[major]) {
                    major = i;
                }
            }
            int minor = (major + 1) % 4;
            for (TAlnCol::iterator it = col.SetCol().begin();
                 it != col.SetCol().end();  ++it) {
                if (it->m_Nt != major) {
                    it->m_Nt = (ENuc) minor;
                }
            }
            if (numAlleles[0] + numAlleles[1] + numAlleles[2] + numAlleles[3] -
                numAlleles[major] == 1) {
                col.SetIsSNP() = eSingletonSNP;
            }
        }

        sort(col.SetCol().begin(), col.SetCol().end());
    }

    return col;
}

/*---------------------------------------------------------------------------*/
// PRE : m_RdBegin & m_RdEnd are set appropriately for m_Pos
// POST: whether this position is (apparently) polymorphic
inline bool CAssm_CI::IsSNPPos(void) const {
    CCol &col = m_Cols[GetPos()-m_Start];
    if (col.GetCol().empty()) {
        **this;//force analysis of this column;
    }
    assert(col.IsSNP() != eUnknown);
    return col.IsSNP() == eSNP  ||  col.IsSNP() == eSingletonSNP;
}

/*---------------------------------------------------------------------------*/
// PRE : m_Pos, m_RdBegin and m_RdEnd initialized, with m_Pos <= the
// chr range
// POST: m_BeginChr and m_EndChr advanced to completely span m_Pos
void CAssm_CI::x_AdvanceRdRange(void) {
    /*
    bool movedBegin =
        (m_RdBegin != m_ReadIdx->end()  &&  (*m_RdBegin)->GetStop() < m_Pos);
    */
    while (m_RdBegin != m_ReadIdx->end()  &&  (*m_RdBegin)->GetStop() < m_Pos){
        ++m_RdBegin;
    }
    /*
    if (movedBegin) {
        m_ChrRangeStart = m_Assm.GetStop();
        m_ChrRangeStop = 0;
        for (TReadIdx::const_iterator it = m_RdBegin;  it < m_RdEnd;  ++it) {
            if ((*it)->GetChr()->GetStop() > m_ChrRangeStop) {
                m_ChrRangeStop = (*it)->GetChr()->GetStop();
            }
            if ((*it)->GetChr()->GetStart() < m_ChrRangeStart) {
                m_ChrRangeStart = (*it)->GetChr()->GetStart();
            }
        }
    }
    */
    while (m_RdEnd != m_ReadIdx->end()  &&  (*m_RdEnd)->GetStart() <= m_Pos) {
        /*
        if ((*m_RdEnd)->GetChr()->GetStop() > m_ChrRangeStop) {
            m_ChrRangeStop = (*m_RdEnd)->GetChr()->GetStop();
        }
        if ((*m_RdEnd)->GetChr()->GetStart() < m_ChrRangeStart) {
            m_ChrRangeStart = (*m_RdEnd)->GetChr()->GetStart();
        }
        */
        ++m_RdEnd;
    }
}

/*---------------------------------------------------------------------------*/
// PRE : initialized
// POST: state "incremented"; whether or not this incrementing "rolled
// over" the counter.  Note this does *NOT* increment in the
// traditional sense as it maintains a constant number of ON bits --
// see code for details
CCombination& CCombination::operator++ (void) {
    //look for rightmost 10 (if doesn't exist, then we've "rolled over")
    bool found = false;
    int i = m_NextPotential10;
    if (i < (int) size()) {
        while (i >= 1) {
            if (!(*this)[i]  &&  (*this)[i-1]) {
                found = true;
                break;
            } else {
                --i;
            }
        }
    }
    if (!found) { //don't bother incrementing counter since we rolled over
        m_RolledOver = true;
        return *this;
    }
    m_NextPotential10 = i-1;
    
    //change to 01
    (*this)[i-1] = false;
    (*this)[i] = true;

    //shuffle bits down to pack tightly: 01+0*$
    int flippedPos = i;
    ++i;
    int j = size() - 1;
    while (i < j) {
        if ((*this)[i]) {
            ++i;
        } else if (!(*this)[j]) {
            --j;
        } else {
            (*this)[i] = true;
            (*this)[j] = false;
            ++i;
            --j;
        }
    }
    if (j < (int) size() - 1) {
        m_NextPotential10 = j+1;
    } else if (!(*this)[j]) {
        m_NextPotential10 = j;
    }
    m_Packed = (i - flippedPos == (int)m_NumOnBits);
    return *this;
}

/*---------------------------------------------------------------------------*/
// PRE : initialized
// POST: next "largest" configuration (if bits ordered in descending order)
void CCombination::AdvanceBeforeNextBiggest(void) {
    int rightmost1 = ((*this)[size()-1]) ? size()-1 : m_NextPotential10;
    while (rightmost1 >= 0  &&  !(*this)[rightmost1]) {
        --rightmost1;
    }

    if (rightmost1 < (int)size()-1) {
        int rightmost0 = size()-1;
        do {
            (*this)[rightmost1] = false;
            (*this)[rightmost0] = true;
            --rightmost1;
            --rightmost0;
        } while (rightmost1 >= 0  &&  (*this)[rightmost1]);
        m_NextPotential10 = rightmost1;
    }
}

/*---------------------------------------------------------------------------*/
// PRE : subset of chromosomes to examine
// POST: probability distribution for # of errors, given the quality scores
void CAssm_CI::CCol::x_CalcErrorDist(const TAlnColSubset &chrs,
                                     TProbDist &errorPrbs) const {
    //when probabilities reach the "small enough" level, we stop calculating (helps with those high-sample depth guys)
    const double kSmallEnough = 1e-10;
    int depth = chrs.size();
    double maxProb = 0;
        
    // loop over all possible values for the # of false alleles
    for (int numFalse = 0;  numFalse <= depth;  ++numFalse) {
        errorPrbs[numFalse] = 0;

        if (numFalse == 0  ||
            errorPrbs[numFalse-1] > kSmallEnough ||
            maxProb == 0) {//only count if we're going to get a non-negligible probability
            if (gsl_sf_choose(depth, numFalse) > 1e6) {
#ifdef _DEBUG
                cerr << "Depth/numFalse too big at #" << GetPos()
                     << "; last prob: " << errorPrbs[numFalse-1]
                     << endl;
#endif
                continue;
            }

            for (CCombination errSet(depth, numFalse);  errSet.good();
                 ++errSet) {
                //for(CCombination::const_iterator it = errSet.begin();  it != errSet.end();  ++it) { cerr << (*it ? 1:0); } cerr << endl;
                double prb = 1;
                
                for (int i = 0;  i < depth;  ++i) {
                    prb *= errSet[i] ?
                        chrs[i]->GetErrProb() : 1 - chrs[i]->GetErrProb();
                }
                
                errorPrbs[numFalse] += prb;
            };

            maxProb = max(maxProb, errorPrbs[numFalse]);
        } //end of bothering to count
    }    
}

/*---------------------------------------------------------------------------*/
// PRE : m_Scores[eMinor]/m_Scores[eMajor] set
// POST: probability of m0 | m stored for all m
void CAssm_CI::CCol::x_CalcObservedFSCondDist(void) {
    TAlnColSubset chrsByNt[4];
    const TAlnCol &col = m_Col;//potentially make this GetCol
    unsigned int depth = 0;
    for (TAlnCol::const_iterator chrI = col.begin();  chrI != col.end();
         ++chrI) {
        if (chrI->m_Nt != eGap  &&  chrI->m_Nt != eN) {
            chrsByNt[chrI->m_Nt].push_back(&(*chrI));
            ++depth;
        }
    }

    //figure out which is the "major" allele (TODO: at some point should figure out way to set ancestral)
    unsigned int major = 0;
    for (unsigned int i = 1;  i < 4;  ++i) {
        if (chrsByNt[i].size() > chrsByNt[major].size()) {
            major = i;
        }
    }
    //merge others into "minor" category
    unsigned int minor = (major + 1) % 4;
    copy(chrsByNt[(major + 2) % 4].begin(), chrsByNt[(major + 2) % 4].end(),
         back_inserter(chrsByNt[minor]));
    copy(chrsByNt[(major + 3) % 4].begin(), chrsByNt[(major + 3) % 4].end(),
         back_inserter(chrsByNt[minor]));
    
    TProbDist prMinus(chrsByNt[minor].size()+1);
    TProbDist prPlus (chrsByNt[major].size()+1);
    x_CalcErrorDist(chrsByNt[minor], prMinus);
    x_CalcErrorDist(chrsByNt[major], prPlus);

    m_prObservedCond.resize(depth+1);
    for (unsigned int m = 0;  m <= depth;  ++m) {
        int trueMinor = m;
        int trueMajor = depth - trueMinor;
        int dm = chrsByNt[minor].size() - m; //aka. m0 - m

        double pr = 0;

        if (dm >= 0) {
            int stop = min(trueMinor, trueMajor - dm);
            for (int i = 0;  i <= stop;  ++i) {
                pr += prMinus[i+dm] * prPlus[i];
            }
        } else {
            int stop = min(trueMinor + dm, trueMajor);
            for (int i = 0;  i <= stop;  ++i) {
                pr += prMinus[i] * prPlus[i-dm];
            }
        }

        m_prObservedCond[m] = pr / gsl_sf_choose(depth, m);
    }
}

/*---------------------------------------------------------------------------*/
// PRE : column loaded
// POST: conditional frequency distribution calculated & output to stream
void CAssm_CI::CCol::WriteFSDist(ostream &o) const {
    if (m_prObservedCond.empty()) {
        const_cast<CCol*>(this)->x_CalcObservedFSCondDist();
    }
    for (TProbDist::const_iterator i = m_prObservedCond.begin();
         i != m_prObservedCond.end();  ++i) {
        o << *i << " ";
    }
}

/*---------------------------------------------------------------------------*/
// PRE : column loaded; stream queued to include ONLY frequency dist
// POST: conditional frequency distribution read from stream
void CAssm_CI::CCol::ReadFSDist(istream &i) {
    double d;
    i >> d;
    while (i) {
        m_prObservedCond.push_back(d);
        i >> d;
    }
}

/*---------------------------------------------------------------------------*/
// PRE : observed # of minor alleles, params for frequency spectrum,
// whether to trust observations (i.e. not use q-scores)
// POST: probability of m0 | frequency spectrum
double CAssm_CI::CCol::GetPrObservedFS(const CFreqSpectrum &fs,
                                       bool trustObs) const {
    if (m_prObservedCond.empty()) {
        const_cast<CCol*>(this)->x_CalcObservedFSCondDist();
    }

    if (trustObs) {
        TAlnColSubset chrsByNt[4];
        const TAlnCol &col = m_Col;//potentially make this GetCol
        unsigned int depth = 0;
        for (TAlnCol::const_iterator chrI = col.begin();  chrI != col.end();
             ++chrI) {
            if (chrI->m_Nt != eGap  &&  chrI->m_Nt != eN) {
                chrsByNt[chrI->m_Nt].push_back(&(*chrI));
                ++depth;
            }
        }

        //figure out which is the "major" allele (TODO: at some point should figure out way to set ancestral)
        unsigned int major = 0;
        for (unsigned int i = 1;  i < 4;  ++i) {
            if (chrsByNt[i].size() > chrsByNt[major].size()) {
                major = i;
            }
        }
        
        return fs(depth-chrsByNt[major].size(), depth);
    }

    double pr = 0;
    //Note: loop *intentionally* runs to Depth -- for symmetry which
    //eases handling of folded spectrum
    for (unsigned int m = 0;  m <= m_prObservedCond.size()-1;  ++m) {
        //Pr(O|T) + Pr(T)
        pr += m_prObservedCond[m] * fs(m, m_prObservedCond.size()-1);
    }

    return pr;
}

/*---------------------------------------------------------------------------*/
// PRE : open input stream containing 2-locus likelihoods in specified format
// POST: table loaded
C2LocusProbs::C2LocusProbs(istream &in, EWhichSites s, ELookupTableFmt fmt) {
    m_WhichSites = s;

    switch (fmt) {
    case eHudson: LoadHudson(in); break;
    case eMcVean: LoadMcVean(in); break;
    case eAuto:
        string errs;
        try { LoadHudson(in); return; } catch (exception &e) {errs = e.what();}
        in.seekg(0);
        try { LoadMcVean(in); return; } catch (exception &e) {
            errs += "\n" + string(e.what());}
        throw runtime_error("Failed to load probs using any known format:\n" +
                            errs);
    }
}

/*---------------------------------------------------------------------------*/
// PRE : table loaded for depth n
// POST: normalization constants calculated for depth n
void C2LocusProbs::x_CalcNormConst(unsigned int n) {
    TProbs &normProbs = m_2LocusProbs.Set(n).m_NormProbs;
    normProbs.clear();
    normProbs.resize(m_RhoValues.size(), log(0));
    const TCfgTable &cfgProbs = m_2LocusProbs[n].m_CfgProbs;

    T2LocusConfig cfg;
    unsigned int start = (m_WhichSites == CAssm_CI::eSegNoSingle ? 2 : 1);
    for (unsigned int c1x = start;  c1x <= n-start;  ++c1x) {
        unsigned int stop =  c1x;
    for (unsigned int cx1 = start;  cx1 <= stop;  ++cx1) {
        if ((c1x==0 && cx1==0)  ||  (c1x==n  &&  cx1==n)  ||
            (c1x==n && cx1==0)) { //DEBUG for norm to singly-seg.
            continue;
        }
    for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
         cfg[e11] <= cx1;  ++cfg[e11]) {
        cfg[e01] = cx1 - cfg[e11];
        cfg[e10] = c1x - cfg[e11];
        cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11];
        
        const TProbs &prbs = cfgProbs[cfg];
        //cerr << "Norm: " << cfg << ": " << prbs[0] << endl;//DEBUG
        for (unsigned int rhoPos = 0;  rhoPos < m_RhoValues.size();  ++rhoPos){
            normProbs[rhoPos] = SumLogged(normProbs[rhoPos], prbs[rhoPos]
                                          + (cfg[e01] != cfg[e10] ? log(2):0));
        }

    }
    }
    }

    //normalization constants for theta adjustments
    TDepthProbs &dp = m_2LocusProbs.Set(n);
    dp.m_A1 = 0;
    for (unsigned int i = 1;  i <= n-1;  ++i) {
        dp.m_A1 += 1.0/i;
    }
    dp.m_SumSegSeg.clear();
    dp.m_SumSegSeg.resize(m_RhoValues.size(), log(0));
    dp.m_SumSegFix.clear();
    dp.m_SumSegFix.resize(m_RhoValues.size(), log(0));

    for (unsigned int c1x = 0;  c1x <= n;  ++c1x) {
        for (unsigned int cx1 = 0;  cx1 <= c1x;  ++cx1) {
            //avoid double fixed
            if ((c1x==0 && cx1==0)  ||  (c1x==n  &&  cx1==n)  ||
                (c1x==n && cx1==0)  ||  (c1x==0  &&  cx1==n)) {
                continue;
            }
            //are we singly-segregating or double-segregating?
            TProbs &summer = (c1x > 0  &&  cx1 == 0) ?
                dp.m_SumSegFix : dp.m_SumSegSeg;

            for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
                 cfg[e11] <= cx1;  ++cfg[e11]) {
                cfg[e01] = cx1 - cfg[e11];
                cfg[e10] = c1x - cfg[e11];
                cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11];

                const TProbs &prbs = cfgProbs[cfg];
                for (unsigned int rhoPos = 0;  rhoPos < m_RhoValues.size();
                     ++rhoPos){
                    summer[rhoPos] = SumLogged
                        (summer[rhoPos], prbs[rhoPos]
                         + (cfg[e01] != cfg[e10] ? log(2):0));
                }
            }
        }
    }
    
}

/*---------------------------------------------------------------------------*/
// PRE : open input stream containing 2-locus likelihoods in format of
// McVean's LDHAT package
// POST: table loaded
void C2LocusProbs::LoadMcVean(istream &in) {
    string line;
    in >> m_MaxDepth;
    getline(in, line);//skip rest of line
    m_Pairtype2Cfg.resize(m_MaxDepth+1);
    TCfgTable &cfgProbs = m_2LocusProbs.Set(m_MaxDepth).m_CfgProbs;
    cfgProbs.clear();
    m_RhoValues.clear();

    {
        int dummy;
        in >> dummy >> m_TableTheta;
        if (!in) {
            throw runtime_error("LoadMcVean: could not find theta in header.");
        }
    }
    {
        unsigned int numRho;
        double maxRho;
        in >> numRho >> maxRho;
        if (!in.good()) {
            throw runtime_error("Failed to load 2-locus probabilities.");
        }
        m_RhoValues.push_back(0);
        for (unsigned int i = 1;  i < numRho-1;  ++i) {
            //weird division to match LDHat
            m_RhoValues.push_back(maxRho*i/(numRho-1));
        }
        m_RhoValues.push_back(maxRho);
    }

    while (getline(in, line).good()) {
        if (line.empty()  ||  line.find("Type") != string::npos) {
            continue;
        }
        T2LocusConfig cfg;
        {
            istringstream iss(line.substr(line.find('#')+1));
            iss >> cfg[e00] >> cfg[e01] >> cfg[e10] >> cfg[e11];
        }
        istringstream iss(line.substr(line.find(':')+1));
        TProbs &prs = cfgProbs.Set(cfg);
        prs.resize(m_RhoValues.size(), log(0));

        T2LocusConfig t;
        t[e00]=cfg[e10]; t[e01]=cfg[e11]; t[e10]=cfg[e00]; t[e11]=cfg[e01];
        TProbs &prsL = cfgProbs.Set(t); prsL.resize(m_RhoValues.size(),log(0));
        t[e00]=cfg[e01]; t[e01]=cfg[e00]; t[e10]=cfg[e11]; t[e11]=cfg[e10];
        TProbs &prsR = cfgProbs.Set(t); prsR.resize(m_RhoValues.size(),log(0));
        t[e00]=cfg[e11]; t[e01]=cfg[e10]; t[e10]=cfg[e01]; t[e11]=cfg[e00];
        TProbs &prsB = cfgProbs.Set(t); prsB.resize(m_RhoValues.size(),log(0));
        /*
        unsigned int multiplicity;
        if (cfg[e00] == cfg[e01]  &&  cfg[e01] == cfg[e10]  &&
            cfg[e10] == cfg[e11]) {
            multiplicity = 1;
        } else if ((cfg[e00] == cfg[e01]  &&  cfg[e10] == cfg[e11])  ||
                   (cfg[e00] == cfg[e11]  &&  cfg[e01] == cfg[e10])  ||
                   (cfg[e00] == cfg[e10]  &&  cfg[e01] == cfg[e11])) {
            multiplicity = 2;
        } else {
            multiplicity = 4;
        }
        */

        double d;
        iss >> d;
        unsigned int rhoPos = 0;
        while (iss) {
            //make probability for unordered configuration
            d += gsl_sf_lnfact(m_MaxDepth) -
                gsl_sf_lnfact(cfg[e00]) -
                gsl_sf_lnfact(cfg[e01]) - 
                gsl_sf_lnfact(cfg[e10]) -
                gsl_sf_lnfact(cfg[e11]);
            //d -= log(multiplicity);
            prs[rhoPos] = prsL[rhoPos] = prsR[rhoPos] = prsB[rhoPos] = d;
            //prs[rhoPos] = SumLogged(prs[rhoPos], d);
            ++rhoPos;
            iss >> d;
        }
        assert(rhoPos == m_RhoValues.size());
    }

    //calculate pr(no seg at either site), if applicable
    try {
        unsigned int n = m_MaxDepth;
        T2LocusConfig cfg;
        TProbs totProb(m_RhoValues.size(), log(0));

        for (unsigned int c1x = 0;  c1x <= n;  ++c1x) {
        for (unsigned int cx1 = 0;  cx1 <= c1x;  ++cx1) {
            //avoid double fixed
            if ((c1x==0 && cx1==0)  ||  (c1x==n  &&  cx1==n)  ||
                (c1x==n && cx1==0)  ||  (c1x==0  &&  cx1==n)) {
                continue;
            }
        for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
             cfg[e11] <= cx1;  ++cfg[e11]) {
            cfg[e01] = cx1 - cfg[e11];
            cfg[e10] = c1x - cfg[e11];
            cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11];

            const TProbs &prbs = cfgProbs[cfg];
            for (unsigned int rhoPos = 0;  rhoPos < m_RhoValues.size();
                 ++rhoPos){
                totProb[rhoPos] = SumLogged
                    (totProb[rhoPos], prbs[rhoPos]
                     + (cfg[e01] != cfg[e10] ? log(2):0));
            }
        }
        }
        }
    
        cfg[e00]=n;
        cfg[e01]=cfg[e10]=cfg[e11]=0;
        TProbs &prs = cfgProbs.Set(cfg);
        prs.resize(m_RhoValues.size(), log(0));
        T2LocusConfig t;
        t[e00]=cfg[e10]; t[e01]=cfg[e11]; t[e10]=cfg[e00]; t[e11]=cfg[e01];
        TProbs &prsL = cfgProbs.Set(t); prsL.resize(m_RhoValues.size(),log(0));
        t[e00]=cfg[e01]; t[e01]=cfg[e00]; t[e10]=cfg[e11]; t[e11]=cfg[e10];
        TProbs &prsR = cfgProbs.Set(t); prsR.resize(m_RhoValues.size(),log(0));
        t[e00]=cfg[e11]; t[e01]=cfg[e10]; t[e10]=cfg[e01]; t[e11]=cfg[e00];
        TProbs &prsB = cfgProbs.Set(t); prsB.resize(m_RhoValues.size(),log(0));
        for (unsigned int rhoPos = 0;  rhoPos < m_RhoValues.size();  ++rhoPos){
            //cout << exp(totProb[rhoPos]) << "\t";
            prs[rhoPos] = prsL[rhoPos] = prsR[rhoPos] = prsB[rhoPos] =
                SubLogged(log(1), totProb[rhoPos])-log(4);

        }
        //cout << endl;
        //exit(0);
    } catch (exception&) {} //presumably just didn't have likelihoods for pairs
                            //pairs with only a single segregating site

    //x_ChangeTheta(0.005, m_MaxDepth);
    x_CalcNormConst(m_MaxDepth);
}

/*---------------------------------------------------------------------------*/
// PRE : open input stream containing 2-locus likelihoods in format of
// Hudson's programs
// POST: table loaded
void C2LocusProbs::LoadHudson(istream &in) {
    unsigned int numPops, numRho;
    string line;

    in >> numPops >> m_MaxDepth >> numRho;
    if (!in.good()) {
        throw runtime_error("LoadHudson: error loading 2-locus probabilities");
    }

    {        
        m_RhoValues.clear();
        getline(in, line);
        istringstream iss(line);
        for (unsigned int i = 1;  i <= numRho;  ++i) {
            double rho;
            iss >> rho;
            if (!iss.good()) {
                throw runtime_error("LoadHudson: error reading rho values.");
            }
            m_RhoValues.push_back(rho);
        }
    }

    m_Pairtype2Cfg.resize(m_MaxDepth+1);
    TCfgTable &cfgProbs = m_2LocusProbs.Set(m_MaxDepth).m_CfgProbs;
    cfgProbs.clear();

    {
        unsigned int rhoPos;
        unsigned int c1x = 0, cx1 = 0;
        in >> ws;
        while (getline(in, line).good()) {
            if (line.empty()) {
                continue;
            }
            istringstream iss(line);
            string s;
            if (line.find("freq:") == 0) {
                iss >> s >> c1x >> cx1;
                rhoPos = 0;
            } else if (line.find("rho:") == 0) {
                T2LocusConfig cfg;
                cfg[e11] = (c1x + cx1 > m_MaxDepth) ? c1x + cx1 - m_MaxDepth : 0;
                double rho, pr;
                iss >> s >> rho >> pr;
                while (!iss.fail()) {
                    cfg[e01] = cx1 - cfg[e11];
                    cfg[e10] = c1x - cfg[e11];
                    cfg[e00] = m_MaxDepth - cfg[e10] - cfg[e01] - cfg[e11];
                    cfgProbs.Set(cfg).push_back(log(pr));
                    /*
                    if (rhoPos == 0) {//DEBUG
                        cerr << "raw@0: " << pr << " <-- " << cfg << endl;
                    }
                    */
                    iss >> pr;
                    ++cfg[e11];
                }
                ++rhoPos;
            } else {
                throw runtime_error("LoadHudson: likelihood file does not conform to Hudson format.");
            }
            in >> ws;
        }
    }

    x_CalcNormConst(m_MaxDepth);
}


/*---------------------------------------------------------------------------*/
// PRE : open input stream containing missing data normalization
// constants in format produced by KickNormConst (n L R <const-rho-0>
// <const-rho-1> ...)
// POST: constants loaded into cache
void C2LocusProbs::LoadMissingNormConsts(istream &in) {
    SMissing key;
    string line;
    double d;

    while (getline(in, line).good()) {
        istringstream iss(line);
        iss >> key.m_N >> key.m_L >> key.m_R;
        TProbs &norms = m_MissingNormConsts.Set(key);
        
        iss >> d;
        while (iss) {
            norms.push_back(d);
            iss >> d;
        }

    }

}

/*---------------------------------------------------------------------------*/
// PRE : new theta value, changed for specified n
// POST: *complete* lookup table (not missing) modified to approximate
// appropriate values for new theta
void C2LocusProbs::x_ChangeTheta(double newTheta, unsigned int n) {
    double a1 = 0;
    for (unsigned int i = 1;  i <= n-1;  ++i) {
        a1 += 1.0/i;
    }

    //probabilities are approximate since assume no back mutation
    double newSegProb = 1 - exp(-newTheta * a1);
    double oldSegProb = 1 - exp(-m_TableTheta * a1);

    TCfgTable &cfgProbs = m_2LocusProbs.Set(n).m_CfgProbs;

    T2LocusConfig cfg;
    TProbs totProb(m_RhoValues.size(), log(0));

    for (unsigned int c1x = 0;  c1x <= n;  ++c1x) {
        for (unsigned int cx1 = 0;  cx1 <= c1x;  ++cx1) {
            //avoid double fixed
            if ((c1x==0 && cx1==0)  ||  (c1x==n  &&  cx1==n)  ||
                (c1x==n && cx1==0)  ||  (c1x==0  &&  cx1==n)) {
                continue;
            }
            //are we singly-segregating or double-segregating?
            double ratio = log
                ((c1x > 0  &&  cx1 == 0) ?
                 newSegProb * (1-newSegProb) / (oldSegProb * (1-oldSegProb)) :
                 newSegProb * newSegProb / (oldSegProb * oldSegProb));

            for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
                 cfg[e11] <= cx1;  ++cfg[e11]) {
                cfg[e01] = cx1 - cfg[e11];
                cfg[e10] = c1x - cfg[e11];
                cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11];

                TProbs &prbs = cfgProbs.Set(cfg);
                for (unsigned int rhoPos = 0;  rhoPos < m_RhoValues.size();
                     ++rhoPos){
                    prbs[rhoPos] += ratio;
                    totProb[rhoPos] = SumLogged
                        (totProb[rhoPos], prbs[rhoPos]
                         + (cfg[e01] != cfg[e10] ? log(2):0));
                }
            }
        }
    }
    
    cfg[e00]=n;
    cfg[e01]=cfg[e10]=cfg[e11]=0;
    TProbs &prs = cfgProbs.Set(cfg);
    prs.resize(m_RhoValues.size(), log(0));
    T2LocusConfig t;
    t[e00]=cfg[e10]; t[e01]=cfg[e11]; t[e10]=cfg[e00]; t[e11]=cfg[e01];
    TProbs &prsL = cfgProbs.Set(t); prsL.resize(m_RhoValues.size(),log(0));
    t[e00]=cfg[e01]; t[e01]=cfg[e00]; t[e10]=cfg[e11]; t[e11]=cfg[e10];
    TProbs &prsR = cfgProbs.Set(t); prsR.resize(m_RhoValues.size(),log(0));
    t[e00]=cfg[e11]; t[e01]=cfg[e10]; t[e10]=cfg[e01]; t[e11]=cfg[e00];
    TProbs &prsB = cfgProbs.Set(t); prsB.resize(m_RhoValues.size(),log(0));
    for (unsigned int rhoPos = 0;  rhoPos < m_RhoValues.size();  ++rhoPos){
        prs[rhoPos] = prsL[rhoPos] = prsR[rhoPos] = prsB[rhoPos] =
            SubLogged(log(1), totProb[rhoPos])-log(4);

    }

}

/*---------------------------------------------------------------------------*/
// PRE : rho
// POST: position of first tabulated rho below this rho
inline unsigned int C2LocusProbs::x_GetFloorRhoPos(double rho) const {
    unsigned int pos = lower_bound(m_RhoValues.begin(),
                                   m_RhoValues.end(), rho) -
        m_RhoValues.begin();
    if (pos > 0  &&  pos < m_RhoValues.size()  &&  rho != m_RhoValues[pos]) {
        --pos; //lower_bound gives first pos can insert
    }

    if (pos >= m_RhoValues.size()) {
        pos = m_RhoValues.size()-1; //this is what Hudson does, which makes sense -- up to a point.  As long as there are relatively few of these guys, its probably ok (the lik. curve is pretty flat at this point anyway).

        //but let's keep things reasonable and cap how much we extrapolate
        if (rho > m_RhoValues.back() * 1.5) {
            throw range_error("Rho value ("+DblToString(rho)+
                              ") greater than tabulated range");
        }
    }

    return pos;
}

/*---------------------------------------------------------------------------*/
// PRE : two locus configuration (4 components), restrictions on how to swap
// POST: single integer representation of configuration (0..x_MaxPairtype)
unsigned int C2LocusProbs::x_Cfg2Pairtype(const T2LocusConfig &cfg,
                                          unsigned int restrict) const {
    assert(cfg[e01] + cfg[e11] > 0  &&  cfg[e01] + cfg[e11] < cfg.N()  &&
           cfg[e10] + cfg[e11] > 0  &&  cfg[e10] + cfg[e11] < cfg.N());
    //with extensive thanks to Gil McVean's LDHAT package
    /*
    T2LocusConfig cfg2(cfg);
    if (cfg2[e10] + cfg2[e11] > cfg2 [e01] + cfg2[e00]) { //if first pos 1>0
        swap(cfg2[e00], cfg2[e10]);
        swap(cfg2[e11], cfg2[e01]);
    }
    if (cfg2[e01] + cfg2[e11] > cfg2 [e10] + cfg2[e00]  || //if second pos 1>0
        (cfg2[e01] + cfg2[e11] == cfg2 [e10] + cfg2[e00]  &&
        cfg2[e00] < cfg2[e01])) {
        swap(cfg2[e00], cfg2[e01]);
        swap(cfg2[e11], cfg2[e10]);
    }

    if (cfg2[e10] + cfg2[e11] == cfg2 [e01] + cfg2[e00]  && //correct extra symmetry
        cfg2[e00] < cfg2[e10]) { 
        swap(cfg2[e00], cfg2[e10]);
        swap(cfg2[e11], cfg2[e01]);
    }

    if (cfg2[e10] < cfg2[e01]  &&  !(restrict & eNoLR)) { //if first < second
        swap(cfg2[e10], cfg2[e01]);
    }
    unsigned int p1 = cfg2[e10] + cfg2[e11];
    unsigned int p2 = cfg2[e01] + cfg2[e11];
    //cerr << "PAIR TYPE: " << p1*(p1-1)*(p1+4)/6 + (p2-1)*(p2+2)/2 + cfg2[e01] << "\t" << p1 << "\t" << p2 << "\t" << cfg2[e01] << endl;
    return p1*(p1-1)*(p1+4)/6 + (p2-1)*(p2+2)/2 + cfg2[e01];
    */

    if (cfg[e10] + cfg[e11] > cfg [e01] + cfg[e00]) { //if first pos 1>0
        if (cfg[e01] + cfg[e11] > cfg [e10] + cfg[e00]) { //if second pos 1>0
            unsigned int p1 = cfg[e01] + cfg[e00];
            unsigned int p2 = cfg[e10] + cfg[e00];
            assert(p1 > 0  &&  p2 > 0);//otherwise not a doubly-polymorphic!
            return (p1 >= p2  ||  (restrict & eNoLR)) ?
                p1*(p1-1)*(p1+4)/6 + (p2-1)*(p2+2)/2 + cfg[e10] :
                p2*(p2-1)*(p2+4)/6 + (p1-1)*(p1+2)/2 + cfg[e01];
        } else {
            unsigned int p1 = cfg[e00] + cfg[e01];
            unsigned int p2 = cfg[e11] + cfg[e01];
            assert(p1 > 0  &&  p2 > 0);//otherwise not a doubly-polymorphic!
            return (p1 >= p2  ||  (restrict & eNoLR)) ?
                p1*(p1-1)*(p1+4)/6 + (p2-1)*(p2+2)/2 + cfg[e11] :
                p2*(p2-1)*(p2+4)/6 + (p1-1)*(p1+2)/2 + cfg[e00];
        }
    } else {
        if (cfg[e01] + cfg[e11] > cfg [e10] + cfg[e00]) { //if second pos 1>0
            unsigned int p1 = cfg[e11] + cfg[e10];
            unsigned int p2 = cfg[e00] + cfg[e10];
            assert(p1 > 0  &&  p2 > 0);//otherwise not a doubly-polymorphic!
            return (p1 >= p2  ||  (restrict & eNoLR)) ?
                p1*(p1-1)*(p1+4)/6 + (p2-1)*(p2+2)/2 + cfg[e00] :
                p2*(p2-1)*(p2+4)/6 + (p1-1)*(p1+2)/2 + cfg[e11];
        } else {
            unsigned int p1 = cfg[e10] + cfg[e11];
            unsigned int p2 = cfg[e01] + cfg[e11];
            assert(p1 > 0  &&  p2 > 0);//otherwise not a doubly-polymorphic!
            return (p1 >= p2  ||  (restrict & eNoLR)) ?
                p1*(p1-1)*(p1+4)/6 + (p2-1)*(p2+2)/2 + cfg[e01] :
                p2*(p2-1)*(p2+4)/6 + (p1-1)*(p1+2)/2 + cfg[e10];
        }
    }

}

/*---------------------------------------------------------------------------*/
// PRE : single integer representation of configuration
// POST: equivalent canonical configuration
const T2LocusConfig&
C2LocusProbs::x_Pairtype2Cfg(unsigned int pt, unsigned int n) const {
    try {
        assert(n < m_Pairtype2Cfg.size());
        return m_Pairtype2Cfg[n][pt];
    } catch(...) {
        T2LocusConfig cfg;
        for (unsigned int c1x = 1;  c1x <= n/2;  ++c1x) {
            for (unsigned int cx1 = 1;  cx1 <= c1x;  ++cx1) {
                for (cfg[e11] = 0;  cfg[e11] <= cx1;  ++cfg[e11]) {
                    cfg[e01] = cx1 - cfg[e11];
                    cfg[e10] = c1x - cfg[e11];
                    cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11];
                    m_Pairtype2Cfg[n].Set(x_Cfg2Pairtype(cfg)) = cfg;
                }
            }
        }
        return m_Pairtype2Cfg[n][pt];
    }
}

/*---------------------------------------------------------------------------*/
// PRE : depth
// POST: maximum single integer representation of configuration
unsigned int C2LocusProbs::x_MaxPairtype(unsigned int n) const {
    n /= 2;
    return n*(n-1)*(n+4)/6 + (n-1)*(n+2)/2 + n;
}

/*---------------------------------------------------------------------------*/
// PRE : raw sample probability at depth m_MaxDepth
// POST: cached probabilities downsampled to depth = n
const C2LocusProbs::TDepthProbs&
C2LocusProbs::x_Downsample(unsigned int n) {
    if (n >= m_MaxDepth) {
        throw overflow_error("Error: supplied two-locus probabilities are too low for sequence depth ("+IntToString(n)+")");
    }
    assert(n > 1);//n==1 is meaningless (and probably the result of a bug)
    const TCfgTable &parentCfgProbs = x_GetDepthProbs(n+1).m_CfgProbs;

    TCfgTable &cfgProbs = m_2LocusProbs.Set(n).m_CfgProbs;

    bool hasFixedCfgs;
    {
        T2LocusConfig fixedCfg;
        fixedCfg[e00] = n+1;
        hasFixedCfgs = parentCfgProbs.Get(fixedCfg);
    }

    T2LocusConfig cfg;
    unsigned int start = hasFixedCfgs ? 0 : 1;
    unsigned int stop = hasFixedCfgs ? n : n-1;

    for (unsigned int c1x = start;  c1x <= stop;  ++c1x) {
        for (unsigned int cx1 = start;  cx1 <= c1x;  ++cx1) {
            for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
                 cfg[e11] <= min(cx1, c1x);  ++cfg[e11]) {
                cfg[e01] = cx1 - cfg[e11];
                cfg[e10] = c1x - cfg[e11];
                cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11];
                //cerr << "n: " << n << " / " << x_Cfg2Pairtype(cfg) << " / " << cfg << endl;

                //assert(x_Cfg2Pairtype(cfg) <= x_MaxPairtype(n));
                TProbs &prs = cfgProbs.Set(cfg);
                prs.resize(m_RhoValues.size(), log(0));

                for (unsigned int i = 0;  i < 4;  ++i) {
                    ++cfg[i];
                    for (unsigned int r = 0;  r < prs.size();  ++r) {
                        prs[r] = SumLogged(prs[r],
                                           log(cfg[i]) - log(n+1) +
                                           parentCfgProbs[cfg][r]);
                    }
                    --cfg[i];
                }
            }
        }
    }
    x_CalcNormConst(n);
#ifdef _DEBUG
    cerr << "Created likelihoods for depth = " << n << endl;
    /*
    for (unsigned int i =0; i < cfgProbs.size();  ++i) {
        cout << i << " -- " << cfgProbs[i][0] << endl;
    }
    */
#endif
    return m_2LocusProbs.Set(n);
}

/*---------------------------------------------------------------------------*/
// PRE : complete (no missing data) & # of chromosomes that should be
// missing on L & R
// POST: # of ways to achieve this amount of missing data
double C2LocusProbs::x_PrObsCondCmpNorm(unsigned int L, unsigned int R,
                                        const T2LocusConfig &cmp) const {
    //note this is not correct (it does not condition on the observation being segregating), but I'm curious to see how it works, practically.
    /*
    return (gsl_sf_lnfact(cmp.N()) -
            gsl_sf_lnfact(L) -
            gsl_sf_lnfact(R) -
            gsl_sf_lnfact(cmp.N()-L-R));
    */
    SWays2Lose key(L, R, cmp);
    try {
        return m_Ways2Lose[key];
    } catch (...) {} //missed cache
    
    double &r = m_Ways2Lose.Set(key);
    r = log(0);
    for (unsigned int L00 = 0;  L00 <= min(L, (unsigned int) cmp[e00]);  ++L00) {
        for (unsigned int L01 = 0;  L01 <= min(L-L00, (unsigned int) cmp[e01]);  ++L01) {
            for (unsigned int L10 = 0;  L10 <= min(L-L00-L01, (unsigned int) cmp[e10]);  
                 ++L10) {
                unsigned int L11 = L-L00-L01-L10;
                if (L11 <= cmp[e11]  &&
                    (cmp[e00]-L00 + cmp[e01]-L01 > 0)  &&
                    (cmp[e10]-L10 + cmp[e11]-L11 > 0)) {

    for (unsigned int R00 = 0;  R00 <= min(R, cmp[e00]-L00);  ++R00) {
        for (unsigned int R01 = 0;  R01 <= min(R-R00, cmp[e01]-L01);  ++R01) {
            for (unsigned int R10 = 0;  R10 <= min(R-R00-R01, cmp[e10]-L10);
                 ++R10) {
                unsigned int R11 = R-R00-R01-R10;
                if (R11 <= cmp[e11]-L11  &&
                    (cmp[e00]-L00-R00 + cmp[e01]-L01-R01 >0  ||  R00+R01 >0) &&
                    (cmp[e10]-L10-R10 + cmp[e11]-L11-R11 >0  ||  R10+R11 >0) &&
                    (cmp[e00]-L00-R00 + cmp[e10]-L10-R10 >0  ||  L00+L10 >0) &&
                    (cmp[e01]-L01-R01 + cmp[e11]-L11-R11 >0  ||  L01+L11 >0)) {
                    r = SumLogged(r,
                        gsl_sf_lnchoose(cmp[e00], L00) +
                        gsl_sf_lnchoose(cmp[e01], L01) +
                        gsl_sf_lnchoose(cmp[e10], L10) +
                        gsl_sf_lnchoose(cmp[e11], L11) +
                        gsl_sf_lnchoose(cmp[e00]-L00, R00) +
                        gsl_sf_lnchoose(cmp[e01]-L01, R01) +
                        gsl_sf_lnchoose(cmp[e10]-L10, R10) +
                        gsl_sf_lnchoose(cmp[e11]-L11, R11)
                                    );
                }
            }}}}
            }
        }
    }
    //cerr << "norm: " << exp(r) << endl;
    /*
    for (unsigned int a = 0;  a <= m;  ++a) { //e0x
        for (unsigned int b = 0;  b <= m-a;  ++b) { //ex1
            for (unsigned int c = 0;  c <= m-a-b;  ++c) {//e1x
                unsigned int d = m-a-b-c; //ex0
                for (unsigned int i = max(0, (int)a-(int)cmp[e01]);
                     i <= min(a, cmp[e00]); ++i) {
                for (unsigned int j = max(0, (int)b-(int)cmp[e11]);
                     j <= min(b, cmp[e01]-(a-i)); ++j) {
                for (unsigned int k = max(0, (int)c-(int)cmp[e10]);
                     k <= min(c, cmp[e11]-(b-j)); ++k) {
                for (unsigned int l = max(0, (int)d-(int)cmp[e00]+(int)i);
                     l <= min(d, cmp[e10]-(c-k)); ++l) {
                    r = SumLogged(r,
                        gsl_sf_lnchoose(cmp[e00], i) +
                        gsl_sf_lnchoose(cmp[e01], a-i) +
                        gsl_sf_lnchoose(cmp[e01]-(a-i), j) +
                        gsl_sf_lnchoose(cmp[e11], b-j) +
                        gsl_sf_lnchoose(cmp[e11]-(b-j), k) +
                        gsl_sf_lnchoose(cmp[e10], c-k) +
                        gsl_sf_lnchoose(cmp[e10]-(c-k), l) +
                        gsl_sf_lnchoose(cmp[e00]-i, d-l)
                                    );
                }
                }
                }
                }
            }
        }
    }
    */
    return r;
}

/*---------------------------------------------------------------------------*/
// PRE : complete (no missing data) & observed (missing data) configuration
// POST: probability of the latter given the former
double C2LocusProbs::x_PrObsCondCmp(const T2LocusConfig &obs,
                                    const T2LocusConfig &cmp) const {
    //cerr << cmp << "\n" << obs << endl;
    double res = log(0);
    if (cmp[e00] < obs[e00]  ||  cmp[e01] < obs[e01]  ||
        cmp[e10] < obs[e10]  ||  cmp[e11] < obs[e11]  ||
        cmp[e00]-obs[e00] + cmp[e01]-obs[e01] < obs[e0x]  ||
        cmp[e10]-obs[e10] + cmp[e11]-obs[e11] < obs[e1x]  ||
        cmp[e00]-obs[e00] + cmp[e10]-obs[e10] < obs[ex0]  ||
        cmp[e01]-obs[e01] + cmp[e11]-obs[e11] < obs[ex1]  ||
        (int)obs[ex0] - (int)(cmp[e10]-obs[e10]) +
        (int)obs[e1x] - (int)(cmp[e11]-obs[e11]) +
        (int)obs[ex1] - (int)(cmp[e01]-obs[e01]) + (int)obs[e0x] < 0  ||

        (int)obs[e1x] - (int)(cmp[e11]-obs[e11]) +
        (int)obs[ex1] - (int)(cmp[e01]-obs[e01]) + (int)obs[e0x] < 0  ||

        (int)obs[ex1] - (int)(cmp[e01]-obs[e01]) + (int)obs[e0x] < 0
) {
        return res; //complete clearly not compatible with observed
    }

    unsigned int iMin = max(0,
                        max((int)obs[e0x]-((int)cmp[e01]-(int)obs[e01]),
                        max((int)obs[ex1]-((int)cmp[e01]-(int)obs[e01])+(int)obs[e0x]-((int)cmp[e11]-(int)obs[e11]),
                            (int)obs[e1x]-((int)cmp[e11]-(int)obs[e11])+(int)obs[ex1]-((int)cmp[e01]-(int)obs[e01])+(int)obs[e0x]-((int)cmp[e10]-(int)obs[e10])
                            )));
    unsigned int iMax = min((int) obs[e0x],
                        min(obs[ex1] - (cmp[e01]-obs[e01]) + obs[e0x],
                        min(obs[e1x] - (cmp[e11]-obs[e11]) + obs[ex1] -
                            (cmp[e01]-obs[e01]) + obs[e0x],
                            obs[ex0] - (cmp[e10]-obs[e10]) + obs[e1x] -
                            (cmp[e11]-obs[e11])+ obs[ex1] -
                            (cmp[e01]-obs[e01]) + obs[e0x]
                            )));

    for (unsigned int i = iMin;  i <= iMax;  ++i) {
        //cerr << i << " (" << iMin << ", " << iMax << ")" << endl;
        assert((int)cmp[e01]-(int)obs[e01] >= (int)obs[e0x]-(int)i);
        // each on a separate line for debugging
        gsl_sf_lnchoose(cmp[e00]-obs[e00], i);
        gsl_sf_lnchoose(cmp[e01]-obs[e01], obs[e0x]-i);
        gsl_sf_lnchoose(cmp[e11]-obs[e11],
                        obs[ex1]-(cmp[e01]-obs[e01] - (obs[e0x]-i)));
        gsl_sf_lnchoose(cmp[e10]-obs[e10],
                        obs[e1x]-(cmp[e11]-obs[e11] - (obs[ex1]-(cmp[e01]-obs[e01] - (obs[e0x]-i)))));

        res = SumLogged(res,
                        gsl_sf_lnchoose(cmp[e00]-obs[e00], i) +
                        gsl_sf_lnchoose(cmp[e01]-obs[e01], obs[e0x]-i) +
                        gsl_sf_lnchoose(cmp[e11]-obs[e11],
                                        obs[ex1]-(cmp[e01]-obs[e01] - (obs[e0x]-i))) +
                        gsl_sf_lnchoose(cmp[e10]-obs[e10],
                                        obs[e1x]-(cmp[e11]-obs[e11] - (obs[ex1]-(cmp[e01]-obs[e01] - (obs[e0x]-i)))))
                        );
    }
    res += gsl_sf_lnchoose(cmp[e00], obs[e00]) +
        gsl_sf_lnchoose(cmp[e01], obs[e01]) +
        gsl_sf_lnchoose(cmp[e10], obs[e10]) +
        gsl_sf_lnchoose(cmp[e11], obs[e11]) -
        (gsl_sf_lnfact(cmp.N()) -
         gsl_sf_lnfact(obs[e0x]+obs[e1x]) -
         gsl_sf_lnfact(obs[ex0]+obs[ex1]) -
         gsl_sf_lnfact(obs.N()-obs.Nmissing()));
    //x_PrObsCondCmpNorm(obs[ex0]+obs[ex1], obs[e0x]+obs[e1x], cmp);
    /* -
        (gsl_sf_lnfact(cmp.N2()) -
         gsl_sf_lnfact(obs[e00]) -
         gsl_sf_lnfact(obs[e01]) -
         gsl_sf_lnfact(obs[e10]) -
         gsl_sf_lnfact(obs[e11]) -
         gsl_sf_lnfact(obs[e0x]) -
         gsl_sf_lnfact(obs[e1x]) -
         gsl_sf_lnfact(obs[ex0]) -
         gsl_sf_lnfact(obs[ex1])
         );*/

    //assert(res <= 0);
    return res;
}

/*---------------------------------------------------------------------------*/
// PRE : configuration of two SNPs (presumably including missing data)
// POST: probability of observing this configuration calculated for
// all rho values in lookup table
void C2LocusProbs::x_CalcPrMissing(CParams tmpP, const T2LocusConfig &obs,
                                   TProbs &probs) const
{
    probs.clear();
    probs.resize(m_RhoValues.size(), log(0));
    T2LocusConfig e; //exact config

    //sum over missing data
    /*
    for (unsigned int i00 = 0;  i00 <= obs[e0x];  ++i00) {
        for (unsigned int i10 = 0;  i10 <= obs[e1x];  ++i10) {
            for (unsigned int j00 = 0;  j00 <= obs[ex0];  ++j00) {
                for (unsigned int j01 = 0;  j01 <= obs[ex1];  ++j01) {
                    e[e00] = obs[e00] + i00 + j00;
                    e[e01] = obs[e01] + (obs[e0x] - i00) + j01;
                    e[e10] = obs[e10] + i10 + (obs[ex0] - j00);
                    e[e11] = obs[e11] + (obs[e1x] - i10) + (obs[ex1] - j01);
    */
    for (unsigned int i00 = 0;  i00 <= obs[e0x] + obs[ex0];  ++i00) {
        for (unsigned int i11 = 0;  i11 <= obs[e1x] + obs[ex1];  ++i11) {
            unsigned int max01 =
                (obs[e0x]-max(0,(int)i00-(int)obs[ex0])) +
                (obs[ex1]-max(0,(int)i11-(int)obs[e1x]));
            for (unsigned int i01 = max(0,(int)obs[e0x]-(int)i00) +
                     max(0, (int)obs[ex1]-(int)i11);  i01 <= max01;  ++i01) {
                e[e00] = obs[e00] + i00;
                e[e11] = obs[e11] + i11;
                e[e01] = obs[e01] + i01;
                e[e10] = obs.N() - e[e00] - e[e11] - e[e01];

                    //cout << e << endl;
                    /*
                    { //DEBUG
                        cerr << e << endl;
                        cerr << "\t" << exp(x_PrObsCondCmp(obs, e)) << endl;
                        //DEBUG -- Hudson/McVean's method
                        cerr << "\t" << exp(gsl_sf_lnfact(obs.N() - obs.Nmissing()) - gsl_sf_lnfact(obs[e00]) - gsl_sf_lnfact(obs[e01]) - gsl_sf_lnfact(obs[e10]) - gsl_sf_lnfact(obs[e11])
                            + gsl_sf_lnfact(e[e00]) + gsl_sf_lnfact(e[e01]) + gsl_sf_lnfact(e[e10]) + gsl_sf_lnfact(e[e11]) - gsl_sf_lnfact(e.N())
                            + gsl_sf_lnchoose(obs[e0x], i00) +
                            + gsl_sf_lnchoose(obs[e1x], i10) +
                            + gsl_sf_lnchoose(obs[ex0], j00) +
                                            + gsl_sf_lnchoose(obs[ex1], j01))
                             << endl;
                    }
                    cerr << GetPrComplete(p, e) << endl;
                    */
                    double condProb = x_PrObsCondCmp(obs, e);
                    /*
                    cerr << e << " --> " << exp(condProb) << "/" 
                         << exp(x_PrObsCondCmpNorm(obs[ex0]+obs[ex1], obs[e0x]+obs[e1x], e)) << " " << exp(GetPrComplete(tmpP.Set(eRho, 0), e)) << endl; //DEBUG
                    */

                    int i = 0;
                    for (TProbs::iterator prbI = probs.begin();
                         prbI != probs.end();  ++prbI) {
                        (*prbI) = SumLogged
                            ((*prbI),
                             GetPrComplete(tmpP.Set(eRho, m_RhoValues[i]), e) +
                             condProb);
                        ++i;
                    }
                    /*  //Hudson's
                    gsl_sf_lnfact(obs.N() - obs.Nmissing()) - gsl_sf_lnfact(obs[e00]) - gsl_sf_lnfact(obs[e01]) - gsl_sf_lnfact(obs[e10]) - gsl_sf_lnfact(obs[e11])
                        + gsl_sf_lnfact(e[e00]) + gsl_sf_lnfact(e[e01]) + gsl_sf_lnfact(e[e10]) + gsl_sf_lnfact(e[e11]) - gsl_sf_lnfact(e.N())
                        + gsl_sf_lnchoose(obs[e0x], i00) +
                        + gsl_sf_lnchoose(obs[e1x], i10) +
                        + gsl_sf_lnchoose(obs[ex0], j00) +
                        + gsl_sf_lnchoose(obs[ex1], j01)
                                        );
                    */

                    //cerr << "pr: " << m_2LocusProbs.GetPr(p, e) << endl;
                        /*
                    loglike = SumLogged(loglike,
                                        x_PrObsCondCmp(obs, e) +
                                        m_2LocusProbs.GetPr(p, e));
                        */
                    //}
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
// PRE : number missing on left, on right & total sample depth
// POST: normalized probabilities cached for all configurations with
// this amount of missing data
void C2LocusProbs::x_CacheNormPrMissing(const CParams &p,
                                        unsigned int L, unsigned int R,
                                        unsigned int n) const {
    T2LocusConfig cfg;
    TProbs norm(m_RhoValues.size(), log(0));
    //TODO: clean up hacky limits on for loops
    //outermost loop should be min(n/2,...) if McVean probs
    //next loop should be min(c1x, n-R-1) if all syms accounted for
    unsigned int start = (m_WhichSites == CAssm_CI::eSegNoSingle ? 2 : 1);
    for (unsigned int c1x = start;  c1x <= n-L-start;  ++c1x) {
        for (unsigned int cx1 = start;  cx1 <= n-R-start;  ++cx1) {
            for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
                 cfg[e11] <= min(n-R-L,min(cx1,c1x));  ++cfg[e11]) {
            for (cfg[e1x] = max(0,(int)R-(int)(n-L-c1x)-(int)cfg[e11]+(int)cx1-(int)min(L,cx1-cfg[e11]));  cfg[e1x] <= min(R,c1x-cfg[e11]);  ++cfg[e1x]) {
            for (cfg[ex1] = max(0,(int)L-(int)(n-R-cx1)-(int)cfg[e11]+(int)c1x-(int)cfg[e1x]);  cfg[ex1] <= min(L,cx1-cfg[e11]);  ++cfg[ex1]) {
            cfg[e01] = cx1 - cfg[e11] - cfg[ex1];
            cfg[e10] = c1x - cfg[e11] - cfg[e1x];
            cfg[e0x] = R-cfg[e1x];
            cfg[ex0] = L-cfg[ex1];
            cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11] - cfg.Nmissing();

            //cout << c1x << "," << cx1 << "\t" << cfg << endl;
            TProbs &probs = m_MissingProbs.Set(cfg);
            x_CalcPrMissing(p, cfg, probs);
            //generate normalization factor to condition on segregating w/ this amount of missing data
            for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size();  ++rhoI) {
                norm[rhoI] = SumLogged(norm[rhoI], probs[rhoI]);
            }
            }}}
        }
    }    


    //now normalize these probs
    for (unsigned int c1x = start;  c1x <= n-L-start;  ++c1x) {
        for (unsigned int cx1 = start;  cx1 <= n-R-start;  ++cx1) {
            for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
                 cfg[e11] <= min(n-R-L,min(cx1,c1x));  ++cfg[e11]) {
            for (cfg[e1x] = max(0,(int)R-(int)(n-L-c1x)-(int)cfg[e11]+(int)cx1-(int)min(L,cx1-cfg[e11]));  cfg[e1x] <= min(R,c1x-cfg[e11]);  ++cfg[e1x]) {
            for (cfg[ex1] = max(0,(int)L-(int)(n-R-cx1)-(int)cfg[e11]+(int)c1x-(int)cfg[e1x]);  cfg[ex1] <= min(L,cx1-cfg[e11]);  ++cfg[ex1]) {
            cfg[e01] = cx1 - cfg[e11] - cfg[ex1];
            cfg[e10] = c1x - cfg[e11] - cfg[e1x];
            cfg[e0x] = R-cfg[e1x];
            cfg[ex0] = L-cfg[ex1];
            cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11] - cfg.Nmissing();

            TProbs &probs = m_MissingProbs.Set(cfg);
            for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size();  ++rhoI) {
                probs[rhoI] -= norm[rhoI];
            }
            //cout << "\t" << probs[0] << endl;

            }}}
        }
    }

    /*
        for (unsigned int rhoPos = 0;  rhoPos < 5;  ++rhoPos){
            cout << "Missing norm: " << exp(norm[rhoPos]) << endl;
        }
    */
}

/*---------------------------------------------------------------------------*/
// PRE : normalization constants for all data with given parameters
// POST: norm constants with singletons subtracted
void C2LocusProbs::x_RemoveSinglesFromNorms(TProbs &norm,
                                            const CParams &p,
                                            unsigned int L, unsigned int R,
                                            unsigned int n) const {
    T2LocusConfig cfg;

    for (unsigned int c1x = 1;  c1x <= n-L-1;  ++c1x) {
        unsigned int stop =  (c1x == 1 ? n-R-1 : 1);
        for (unsigned int cx1 = 1;  cx1 <= stop;  ++cx1) {
            for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
                 cfg[e11] <= min(n-R-L,min(cx1,c1x));  ++cfg[e11]) {
            for (cfg[e1x] = max(0,(int)R-(int)(n-L-c1x)-(int)cfg[e11]+(int)cx1-(int)min(L,cx1-cfg[e11]));  cfg[e1x] <= min(R,c1x-cfg[e11]);  ++cfg[e1x]) {
            for (cfg[ex1] = max(0,(int)L-(int)(n-R-cx1)-(int)cfg[e11]+(int)c1x-(int)cfg[e1x]);  cfg[ex1] <= min(L,cx1-cfg[e11]);  ++cfg[ex1]) {
            cfg[e01] = cx1 - cfg[e11] - cfg[ex1];
            cfg[e10] = c1x - cfg[e11] - cfg[e1x];
            cfg[e0x] = R-cfg[e1x];
            cfg[ex0] = L-cfg[ex1];
            cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11] - cfg.Nmissing();

            //cout << c1x << "," << cx1 << "\t" << cfg << endl;
            TProbs probs;
            x_CalcPrMissing(p, cfg, probs);

            for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size();  ++rhoI) {
                norm[rhoI] = SubLogged(norm[rhoI], probs[rhoI]);
            }
            }}}
        }
    }    
}

/*---------------------------------------------------------------------------*/
// PRE : parameters; configuration of two SNPs (potentially including
// missing data)
// POST: probability of observing this configuration, conditional on
// both sites being segregating & on the given amount of missing data
double C2LocusProbs::GetPrMissing(const CParams& p,
                                  const T2LocusConfig &obs) const
{
    /* DELETE or UNCOMMENT if keep running all sites through
    assert(obs[e00] + obs[e01] + obs[e0x] > 0  &&
           obs[e10] + obs[e11] + obs[e1x] > 0  &&
           obs[e00] + obs[e10] + obs[ex0] > 0  &&
           obs[e01] + obs[e11] + obs[ex1] > 0);
    */
    if (obs.Nmissing() == 0) {
        return GetPrComplete(p, obs);
    }

    const TProbs *probs = m_MissingProbs.Get(obs);//*canon
    if (!probs) { //missed cache so need to calc
        /*
        try {
            const TProbs &norm = m_MissingNormConsts[SMissing(L, R, obs.N())];
        */
        TProbs &pr = m_MissingProbs.Set(obs);
        x_CalcPrMissing(p, obs, pr);
        if (m_WhichSites != CAssm_CI::eAll) {
            const TProbs *norm =
                m_MissingNormConsts.Get(SMissing(obs.Lmissing(),
                                                 obs.Rmissing(), obs.N()));
            if (!norm) {
#ifdef _DEBUG
                cerr << "Generating norm constants for: " << obs.Lmissing()
                     << "\t" << obs.Rmissing() << "\t" << obs.N() << endl;
#endif
                x_CacheNormPrMissing(p, obs.Lmissing(),
                                     obs.Rmissing(), obs.N());
                probs = m_MissingProbs.Get(obs);
            } else if (m_WhichSites == CAssm_CI::eSegNoSingle) {
                TProbs newNorm(*norm);
                x_RemoveSinglesFromNorms(newNorm, p, obs.Lmissing(),
                                         obs.Rmissing(), obs.N());
                for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size(); ++rhoI){
                    pr[rhoI] -= newNorm[rhoI];
                }
            } else {
                for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size(); ++rhoI){
                    pr[rhoI] -= (*norm)[rhoI];
                }
            }
        }
            /*
            if (m_SingletonOpt == eNoSingle) {
                TProbs newNorm(norm);
                x_RemoveSinglesFromNorms(newNorm, p, L, R, obs.N());
                for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size(); ++rhoI){
                    pr[rhoI] -= newNorm[rhoI];
                }
            } else {
                for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size(); ++rhoI){
                    pr[rhoI] -= norm[rhoI];
                }
            }
            */
            probs = &pr;
            /*
        } catch(...) { //missed norm. cache so need to calc that as well
#ifdef _DEBUG
            cerr << "Calculating norms for n==" << obs.N() << "..." << endl;
#endif
            x_CacheNormPrMissing(p, L, R, obs.N());
            probs = m_MissingProbs.Get(*canon);
        }
            */
    }

    unsigned int rhoPos = x_GetFloorRhoPos(p.Get(eRho));
    if (rhoPos == m_RhoValues.size()-1) {
        return (*probs)[rhoPos];
    } else {
        return x_Interpolate(m_RhoValues[rhoPos], m_RhoValues[rhoPos+1],
                             (*probs)[rhoPos], (*probs)[rhoPos+1],
                             p.Get(eRho));
    }
}

/*---------------------------------------------------------------------------*/
// PRE : depth
// POST: 
void C2LocusProbs::KickNormConst(unsigned int n) const {
    T2LocusConfig cfg;

    for (unsigned int c1x = 1;  c1x <= n/2;  ++c1x) {
        for (unsigned int cx1 = 1;  cx1 <= c1x;  ++cx1) {
            for (cfg[e11] = 0;  cfg[e11] <= cx1;  ++cfg[e11]) {
                
                cfg[e01] = cx1 - cfg[e11];
                cfg[e10] = c1x - cfg[e11];
                cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11] - cfg.Nmissing();


                cfg[e00] = cfg[e01] = cfg[e10] = cfg[e11] = 10;

                cout << cfg << endl;
                for (unsigned int l = 0;  l < n-2;  ++l) {
                    for (unsigned int r = 0; r < n-l;  ++r) {
                        x_PrObsCondCmpNorm(l, r, cfg);
                    }
                }
                cout << "finished 1" << endl;
                for (unsigned int l = 0;  l < n-2;  ++l) {
                    for (unsigned int r = 0; r < n-l;  ++r) {
                        x_PrObsCondCmpNorm(l, r, cfg);
                    }
                }
                cout << "finished 2" << endl;
                exit(0);
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
// PRE : depth
// POST: 
void C2LocusProbs::KickCacheMissing(unsigned int n, const CParams &p) const {
    T2LocusConfig cfg;        
    TProbs probs;
            
    {

        //TODO: clean up hacky limits on for loops
        unsigned int cnt = 0;
        /*
        unsigned int L = 10;
        unsigned int R = 10;
        {{
        */
        for (unsigned int L = 0;  L <= n-2;  ++L) {
        for (unsigned int R = 0;  R <= min(L,n-max((unsigned int)2,L));  ++R) {
    TProbs norm(m_RhoValues.size(), log(0));

    /*
    for (unsigned int c1x = 1;  c1x <= min(n/2, n-L-1);  ++c1x) {
        for (unsigned int cx1 = 1;  cx1 <= min(c1x, n-R-1);  ++cx1) {
            for (cfg[e11] = 0;  cfg[e11] <= min(n-R-L,cx1);  ++cfg[e11]) {
    */
    for (unsigned int c1x = 1;  c1x <= n-L-1;  ++c1x) {
        for (unsigned int cx1 = 1;  cx1 <= n-R-1;  ++cx1) {
            for (cfg[e11] = (c1x + cx1 > n) ? c1x + cx1 - n : 0;
                 cfg[e11] <= min(n-R-L,min(cx1,c1x));  ++cfg[e11]) {
                for (cfg[e1x] = max(0,(int)R-(int)(n-L-c1x)-(int)cfg[e11]+(int)cx1-(int)min(L,cx1-cfg[e11]));  cfg[e1x] <= min(R,c1x-cfg[e11]);  ++cfg[e1x]) {
                for (cfg[ex1] = max(0,(int)L-(int)(n-R-cx1)-(int)cfg[e11]+(int)c1x-(int)cfg[e1x]);  cfg[ex1] <= min(L,cx1-cfg[e11]);  ++cfg[ex1]) {
            cfg[e01] = cx1 - cfg[e11] - cfg[ex1];
            cfg[e10] = c1x - cfg[e11] - cfg[e1x];
            cfg[e0x] = R-cfg[e1x];
            cfg[ex0] = L-cfg[ex1];
            cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11] - cfg.Nmissing();

            x_CalcPrMissing(p, cfg, probs);
            //generate normalization factor to condition on segregating w/ this amount of missing data
            for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size();  ++rhoI) {
                norm[rhoI] = SumLogged(norm[rhoI], probs[rhoI]);
            }


            ++cnt;//cout << c1x << "," << cx1 << "\t" << cfg << endl;
            }}}
        }
    }
        //cout << cnt << endl;
        cout << n << "\t" << L << "\t" << R;
        for(unsigned int rhoI = 0;  rhoI < m_RhoValues.size();  ++rhoI) {
            //cout << n << "\t" << L << "\t" << R << "\t" << m_RhoValues[rhoI] << "\t" << norm[rhoI] << endl;
            cout << "\t" << norm[rhoI];
        }
        cout << endl;
    }}
    return;
    }



    for (unsigned int c1x = 1;  c1x <= n/2;  ++c1x) {
        for (unsigned int cx1 = 1;  cx1 <= c1x;  ++cx1) {
            for (cfg[e11] = 0;  cfg[e11] <= cx1;  ++cfg[e11]) {
            for (cfg[e1x] = 0;  cfg[e1x] <= c1x-cfg[e11];  ++cfg[e1x]) {
                for (cfg[ex1] = 0;  cfg[ex1] <= cx1-cfg[e11];  ++cfg[ex1]) {
  
            cfg[e01] = cx1 - cfg[e11] - cfg[ex1];
            cfg[e10] = c1x - cfg[e11] - cfg[e1x];

            for (cfg[e0x] = 0;  (int)cfg[e0x] <= (int)(n-c1x-cx1+cfg[e11])-(cfg[e10]==0?1:0);  ++cfg[e0x]) {
                for (cfg[ex0] = 0;  (int)cfg[ex0] <= (int)(n-c1x-cx1+cfg[e11]-cfg[e0x])-(cfg[e01]+cfg[e0x]==0?1:0); ++cfg[ex0]) {
            
            /*
            for (cfg[e11] = 0;  cfg[e11] <= cx1;  ++cfg[e11]) {

                for (cfg[e1x] = 0;  cfg[e1x] <= c1x-cfg[e11];
                     ++cfg[e1x]) {
                for (cfg[ex1] = 0;  cfg[ex1] <= min(n-cfg[e11]-cfg[e1x]-1, cx1-cfg[e11]);
                     ++cfg[ex1]) {

                    for (cfg[e0x] = 0;  cfg[e0x] <= min(5*n-2,min(n-cfg[e11]-cfg[e1x]-cfg[ex1]-1,n-c1x)); ++cfg[e0x]) {
                        for (cfg[ex0] = 0;  cfg[ex0] <= min(n-2,min(n-cfg[e11]-cfg[e1x]-cfg[ex1]-cfg[e0x]-1,n-cx1)); ++cfg[ex0]) {


                cfg[e01] = cx1 - cfg[e11] - cfg[ex1];
                cfg[e10] = c1x - cfg[e11] - cfg[e1x];
*/
                cfg[e00] = n - cfg[e10] - cfg[e01] - cfg[e11] - cfg.Nmissing();
                if (!(cfg[e00] < n  && cfg[e01] < n && cfg[e10] < n  && cfg[e11] < n && cfg[e0x] < n && cfg[ex0] < n  &&  cfg[e1x] < n && cfg[ex1] < n)) {
                    cout << c1x << "," << cx1 << "\t" << cfg << endl;
                    //assert(cfg[e00] < n  && cfg[e01] < n && cfg[e10] < n  && cfg[e11] < n && cfg[e0x] < n && cfg[ex0] < n  &&  cfg[e1x] < n && cfg[ex1] < n);
                }                    
                if (!(cfg[e01] + cfg[e11] + cfg[ex1] > 0  &&
                      cfg[e00] + cfg[e10] + cfg[ex0] > 0  &&
                      cfg[e10] + cfg[e11] + cfg[e1x] > 0  &&
                      cfg[e00] + cfg[e01] + cfg[e0x] > 0)) {
                    cout << c1x << "," << cx1 << "\t" << cfg << endl;
                    //assert(cfg[e00] < n  && cfg[e01] < n && cfg[e10] < n  && cfg[e11] < n && cfg[e0x] < n && cfg[ex0] < n  &&  cfg[e1x] < n && cfg[ex1] < n);
                }                    
                //if (cfg.Nmissing() == n) {
                if (cfg[e0x]+cfg[e1x] == 3  &&
                    cfg[ex0]+cfg[ex1] == n/4) {
                    cout << c1x << "," << cx1 << "\t" << cfg << endl;
                    CParams p(eNone);
                    p.Set(eRho, 0);
                    //GetPrMissing(p,cfg);
                }

                }
                }
                }
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
// PRE : parameters; UNLABELED configuration of two SNPs (potentially including
// missing data)
// POST: probability of observing this configuration (or flipping the
// 0/1 labels on the L or R or both), conditional on both sites being
// segregating & on the given amount of missing data
double C2LocusProbs::GetFoldedPr(const CParams& p,
                                 const T2LocusConfig &cfg) const {
#ifdef FOLD_DEBUG
        cout << "\t" << cfg << endl;
#endif
    if (cfg[e00] == cfg[e01]  &&  cfg[e01] == cfg[e10]  &&
        cfg[e10] == cfg[e11]  &&
        cfg[e0x] == cfg[e1x]  &&  cfg[ex0] == cfg[ex1]) {
        return GetPrMissing(p, cfg);

    } else if ((cfg[e00] == cfg[e01]  &&  cfg[e10] == cfg[e11]  &&
                cfg[ex0] == cfg[ex1])  ||
               (cfg[e00] == cfg[e11]  &&  cfg[e01] == cfg[e10]  &&
                cfg[ex0] == cfg[ex1]  &&  cfg[e0x]  == cfg[e1x])) {
        //FIXME -- BELOW ONLY WORKS WITH MCVEAN/SYMMETRIC PROBS
        return log(2) + GetPrMissing(p, cfg);
        //swap L
        T2LocusConfig t;
        t[e00]=cfg[e10]; t[e01]=cfg[e11]; t[e10]=cfg[e00]; t[e11]=cfg[e01];
        t[e0x]=cfg[e1x]; t[e1x]=cfg[e0x]; t[ex0]=cfg[ex0]; t[ex1]=cfg[ex1];
#ifdef FOLD_DEBUG
        cout << "\t" << t << endl;
#endif
        return SumLogged(GetPrMissing(p, cfg), GetPrMissing(p, t));

    } else if (cfg[e00] == cfg[e10]  &&  cfg[e01] == cfg[e11]  &&
               cfg[e0x] == cfg[e1x]) {
        return log(2) + GetPrMissing(p, cfg);
        //swap R
        T2LocusConfig t;
        t[e00]=cfg[e01]; t[e01]=cfg[e00]; t[e10]=cfg[e11]; t[e11]=cfg[e10];
        t[e0x]=cfg[e0x]; t[e1x]=cfg[e1x]; t[ex0]=cfg[ex1]; t[ex1]=cfg[ex0];
#ifdef FOLD_DEBUG
        cout << "\t" << t << endl;
#endif
        return SumLogged(GetPrMissing(p, cfg), GetPrMissing(p, t));

    } else {
        //swap all
        return log(4) + GetPrMissing(p, cfg);

        T2LocusConfig t;
        t[e00]=cfg[e10]; t[e01]=cfg[e11]; t[e10]=cfg[e00]; t[e11]=cfg[e01];
        t[e0x]=cfg[e1x]; t[e1x]=cfg[e0x]; t[ex0]=cfg[ex0]; t[ex1]=cfg[ex1];
#ifdef FOLD_DEBUG
        cout << "\t" << t << endl;
#endif
        double swapL = GetPrMissing(p, t);
        t[e00]=cfg[e01]; t[e01]=cfg[e00]; t[e10]=cfg[e11]; t[e11]=cfg[e10];
        t[e0x]=cfg[e0x]; t[e1x]=cfg[e1x]; t[ex0]=cfg[ex1]; t[ex1]=cfg[ex0];
#ifdef FOLD_DEBUG
        cout << "\t" << t << endl;
#endif
        double swapR = GetPrMissing(p, t);
        t[e00]=cfg[e11]; t[e01]=cfg[e10]; t[e10]=cfg[e01]; t[e11]=cfg[e00];
        t[e0x]=cfg[e1x]; t[e1x]=cfg[e0x]; t[ex0]=cfg[ex1]; t[ex1]=cfg[ex0];
#ifdef FOLD_DEBUG
        cout << "\t" << t << endl;
#endif
        double swapB = GetPrMissing(p, t);
        return SumLogged(SumLogged(SumLogged(GetPrMissing(p, cfg),
                                             swapL),
                                   swapR),
                         swapB);
    }
}

/*---------------------------------------------------------------------------*/
// PRE : configuration
// POST: configuration swapped around to *folded* "canonical" form
void x_Canonicalize(T2LocusConfig &cfg) {
    if (cfg[e10] + cfg[e11] + cfg[e1x] >
        cfg[e01] + cfg[e00] + cfg[e0x]) { //if first pos 1>0
        swap(cfg[e00], cfg[e10]);
        swap(cfg[e11], cfg[e01]);
        swap(cfg[e1x], cfg[e0x]);
    }
    if (cfg[e01] + cfg[e11] + cfg[ex1] >
        cfg[e10] + cfg[e00] + cfg[ex0] || //if second pos 1>0
        (cfg[e01] + cfg[e11] + cfg[ex1] == cfg [e10] + cfg[e00] + cfg[ex0] &&
         cfg[e00] < cfg[e01])) {
        swap(cfg[e00], cfg[e01]);
        swap(cfg[e11], cfg[e10]);
        swap(cfg[ex1], cfg[ex0]);
    }

    if (cfg[e10] + cfg[e11] + cfg[e1x] == cfg [e01] + cfg[e00] + cfg[e0x] && 
        cfg[e00] < cfg[e10]) { //correct extra symmetry
        swap(cfg[e00], cfg[e10]);
        swap(cfg[e11], cfg[e01]);
        swap(cfg[e1x], cfg[e0x]);
    }

    if (cfg[e10] + cfg[e1x] < cfg[e01] + cfg[ex1]) { //if first < second
        swap(cfg[e10], cfg[e01]);
        swap(cfg[e1x], cfg[ex1]);
        swap(cfg[e0x], cfg[ex0]);
    }
}

/*---------------------------------------------------------------------------*/
// The following structs & classes are used only in
// CPIIMmodel::x_CalcObsProbCfgs (below)
struct SAlleleQ {
    unsigned int m_Pos;
    double m_ErrProb;
    bool m_Left;
    E2LocusCat m_Cat;
    SAlleleQ *m_Pair;
    SAlleleQ(void) { m_Pair = NULL; }
};
class CQualCfg : public vector<SAlleleQ*> {
public:
    CQualCfg(const TAlnCol &col1, const TAlnCol &col2) {
        ENuc anc1 = col1.front().m_Nt; //FIXME to be real anc.(?)
        ENuc anc2 = col2.front().m_Nt; 
        TAlnCol::const_iterator c1 = col1.begin(), c2 = col2.begin();
        while (c1 != col1.end()  &&  c2 != col2.end()) {
            if (*c1 < *c2) {
                x_AddChr((c1->m_Nt == anc1) ? e0x : e1x, c1->GetErrProb());
                ++c1;
            } else if (*c2 < *c1) {
                x_AddChr((c2->m_Nt == anc2) ? ex0 : ex1, c2->GetErrProb());
                ++c2;
            } else {
                if (c1->m_Nt == anc1) {
                    if (c2->m_Nt == anc2) {
                        x_AddChr(e00, c1->GetErrProb(), c2->GetErrProb());
                    } else {
                        x_AddChr(e01, c1->GetErrProb(), c2->GetErrProb());
                    }
                } else {
                    if (c2->m_Nt == anc2) {
                        x_AddChr(e10, c1->GetErrProb(), c2->GetErrProb());
                    } else {
                        x_AddChr(e11, c1->GetErrProb(), c2->GetErrProb());
                    }
                }
                ++c1;
                ++c2;
            }
        }
        while (c1 != col1.end()) {
            x_AddChr((c1->m_Nt == anc1) ? e0x : e1x, c1->GetErrProb());
            ++c1;
        }
        while (c2 != col2.end()) {
            x_AddChr((c2->m_Nt == anc2) ? ex0 : ex1, c2->GetErrProb());
            ++c2;
        }        

        Sort(); //sort in descending order of error prob
#ifdef ERR_DEBUG
        for (const_iterator i = begin();  i != end();  ++i) {
            cerr << ((*i)->m_Left ? "L":"R") << "\t"
                 << (*i)->m_Cat;
        }
        cerr << endl;
#endif
    }

    ~CQualCfg(void) {
        for (iterator it = begin();  it != end();  ++it) {
            delete *it;
            *it = NULL;
        }
    }
    const T2LocusConfig& GetObservedCfg(void) const { return m_ObsCfg; }

    void Sort(void) {
        sort(begin(), end(), CCmp());
        for (unsigned int i = 0;  i < size();  ++i) {
            (*this)[i]->m_Pos = i;
        }
    }
    bool ApplyFilters(void);

protected:
    void x_AddChr(E2LocusCat cat, double errProb) {
        assert(cat == e0x  ||  cat == e1x  ||  cat == ex0  ||  cat == ex1);
        SAlleleQ *chr = new SAlleleQ;
        push_back(chr);
        chr->m_Cat = cat;
        chr->m_Left = (cat == e0x || cat == e1x);
        chr->m_ErrProb = errProb;
        ++m_ObsCfg[cat];
    }
    void x_AddChr(E2LocusCat cat, double errProbL, double errProbR) {
        assert(cat == e00  ||  cat == e01  ||  cat == e10  ||  cat == e11);
        SAlleleQ *chrL = new SAlleleQ;
        SAlleleQ *chrR = new SAlleleQ;
        push_back(chrL);
        push_back(chrR);
        chrL->m_Pair = chrR;
        chrR->m_Pair = chrL;
        chrL->m_Cat = chrR->m_Cat = cat;
        chrL->m_Left = true;
        chrL->m_ErrProb = errProbL;
        chrR->m_Left = false;
        chrR->m_ErrProb = errProbR;
        ++m_ObsCfg[cat];
    }

    class CCmp {
    public:
        bool operator() (const SAlleleQ* a, const SAlleleQ* b) const {
            if (a->m_ErrProb > b->m_ErrProb) {
                return true;
            } else if (a->m_ErrProb < b->m_ErrProb) {
                return false;
            } else if (IsMissing(a->m_Cat)  &&  !IsMissing(b->m_Cat)) {
                return true;
            } else if (!IsMissing(a->m_Cat)  &&  IsMissing(b->m_Cat)) {
                return false;
            } else {
                return a->m_Cat < b->m_Cat;
            }
        }
    };

private:
    T2LocusConfig m_ObsCfg;
};

/*---------------------------------------------------------------------------*/
// PRE : alleles should be initialized
// POST: ad hoc filtering to get rid of excessive #s of bad bases
bool CQualCfg::ApplyFilters(void) {
    if (size() < CAssm_CI::sm_MaxBadBases) { //okay by default
        return true;
    }

    unsigned int firstGood;
    for (firstGood = 0;  firstGood < size()  &&
             (*this)[firstGood]->m_ErrProb > CAssm_CI::sm_ErrCutoff;
         ++firstGood) {}

    if (firstGood <= CAssm_CI::sm_MaxBadBases) { //we're ok
        return true;
    }

    //too many bad 'uns..
    unsigned int numToDelete = firstGood - CAssm_CI::sm_MaxBadBases;
    if (size() - numToDelete < 2) {//not redeemable
        cerr << "not redeemable 1" << endl;
        return false;
    }
#ifdef ERR_FILT_DEBUG
    cerr << "BEFO --> ";
    for (const_iterator i = begin();  i != end();  ++i) {
        cerr << (*i)->m_ErrProb << "\t";
    }
    cerr << endl;
#endif
    for (const_iterator it = begin();  it < begin()+numToDelete;  ++it) {
        if ((*it)->m_Pair) {
            (*it)->m_Pair->m_Pair = NULL;
            switch((*it)->m_Pair->m_Cat) {
            case e00: (*it)->m_Pair->m_Cat = (*it)->m_Pair->m_Left ? e0x :ex0;
                break;
            case e01: (*it)->m_Pair->m_Cat = (*it)->m_Pair->m_Left ? e0x :ex1;
                break;
            case e10: (*it)->m_Pair->m_Cat = (*it)->m_Pair->m_Left ? e1x :ex0;
                break;
            case e11: (*it)->m_Pair->m_Cat = (*it)->m_Pair->m_Left ? e1x :ex1;
                break;
            default: assert(false);
            }
        }
        delete *it;
    }
    erase(begin(), begin()+numToDelete);

    //renumber & reset m_ObsCfg
    m_ObsCfg.clear();
    unsigned int c = 0;
    for (const_iterator it = begin();  it != end();  ++it, ++c) {
        (*it)->m_Pos = c;
        if (!(*it)->m_Pair  ||
            (*it)->m_Pos < (*it)->m_Pair->m_Pos) {
            ++m_ObsCfg[(*it)->m_Cat];
        }
    }
#ifdef ERR_FILT_DEBUG
    cerr << "AFTA --> ";
    for (const_iterator i = begin();  i != end();  ++i) {
        cerr << (*i)->m_ErrProb << "\t";
    }
    cerr << endl;
#endif
    //make sure we still have a depth >= 2 at L and R
    unsigned fullChr = m_ObsCfg[e00] + m_ObsCfg[e01] +
        m_ObsCfg[e10] + m_ObsCfg[e11];
    return (fullChr + m_ObsCfg[e0x] + m_ObsCfg[e1x] >= 2  &&
            fullChr + m_ObsCfg[ex0] + m_ObsCfg[ex1] >= 2);
}

/*---------------------------------------------------------------------------*/
// PRE : two columns (should each be apparently polymorphic)
// POST: SNPPair populated with all reasonably probable configurations
// of these two columns
void CPIIMmodel::x_CalcObsProbCfgs(const CAssm_CI &it1, const CAssm_CI &it2,
                                   SSNPPair &snpPair) {
    typedef map<T2LocusConfig, double> TProbCfgMap;
    TProbCfgMap cfgMap;

    assert(it1.good()  &&  it2.good()  &&  it2.GetPos() > it1.GetPos());
    snpPair.m_Dist = it2.GetPos() - it1.GetPos();
    const TAlnCol &col1 = it1->GetCol();
    const TAlnCol &col2 = it2->GetCol();

    CQualCfg qDat(col1, col2);
    if (!qDat.ApplyFilters()) {
        return;
    }

    unsigned int numAlleles = qDat.size();

    //count number of bases w/ Pr(err) > 0.5
    unsigned int numErrAboveHalf = 0;
    for (unsigned int i = 0;  i < qDat.size();  ++i) {
        if (qDat[i]->m_ErrProb > 0.5) {
            ++numErrAboveHalf;
        }
    }

    //when probabilities reach the "small enough" level, we stop calculating
    double maxProb = 1;
    for (unsigned int i = 0;  i < numAlleles;  ++i) {
        maxProb *= max(qDat[i]->m_ErrProb, 1 - qDat[i]->m_ErrProb);
    }
#ifdef ERR_DEBUG
    cerr << "max prob (small 'nuf): " << maxProb << endl;
#endif
    double smallEnough = maxProb * sm_SmallEnough;

    // loop over all possible values for the # of false alleles
    bool done = false;
    for (unsigned int numFalse = 0;  numFalse <= numAlleles  &&  !done;
         ++numFalse) {
        for (CCombination errSet(numAlleles, numFalse);  errSet.good();
             ++errSet) {
            double prb = 1;
            T2LocusConfig cfg = qDat.GetObservedCfg();
            //cerr << errSet << endl;

            for (unsigned int i = 0;  i < numAlleles;  ++i) {
                if (errSet[i]) {
                    if (!qDat[i]->m_Pair) {
                        --cfg[qDat[i]->m_Cat];
                        switch (qDat[i]->m_Cat) {
                        case e0x: ++cfg[e1x]; break;
                        case e1x: ++cfg[e0x]; break;
                        case ex0: ++cfg[ex1]; break;
                        case ex1: ++cfg[ex0]; break;
                        default: assert(false);
                        }
                    } else if (errSet[qDat[i]->m_Pair->m_Pos]) { //dual errors
                        if (qDat[i]->m_Pos < qDat[i]->m_Pair->m_Pos) {
                            --cfg[qDat[i]->m_Cat];
                            switch (qDat[i]->m_Cat) {
                            case e00: ++cfg[e11]; break;
                            case e01: ++cfg[e10]; break;
                            case e10: ++cfg[e01]; break;
                            case e11: ++cfg[e00]; break;
                            default: assert(false);
                            }
                        }
                    } else { //only this is an error
                        --cfg[qDat[i]->m_Cat];
                        switch (qDat[i]->m_Cat) {
                        case e00: ++cfg[qDat[i]->m_Left ? e10:e01]; break;
                        case e01: ++cfg[qDat[i]->m_Left ? e11:e00]; break;
                        case e10: ++cfg[qDat[i]->m_Left ? e00:e11]; break;
                        case e11: ++cfg[qDat[i]->m_Left ? e01:e10]; break;
                        default: assert(false);
                        }
                    }
                    prb *= qDat[i]->m_ErrProb;
                } else {
                    prb *= 1 - qDat[i]->m_ErrProb;
                }
            }
#ifdef ERR_DEBUG
            cerr << prb << " / " << exp(cfg.LogNumOrderedCfgs()) << endl;
#endif
            if (prb <= smallEnough) {
                if (errSet.IsPacked()) {
                    //if small enough in first cfg for this numFalse, then
                    //we're completely done
                    if (errSet[0]  &&  numFalse > numErrAboveHalf) {
                        done = true;
                    }
                    break; //just done with this numFalse
                } else {
                    errSet.AdvanceBeforeNextBiggest();
                }
            } else {
                x_Canonicalize(cfg);
                //switch to ordered probabilities (required for this
                //procedure since qualities must be assigned to a
                //particular ordering of chromosomes)
                prb = log(prb) - cfg.LogNumOrderedCfgs();
                //save configuration & probability
                TProbCfgMap::iterator cfgI = cfgMap.find(cfg);
                if (cfgI == cfgMap.end()) {
                    cfgMap[cfg] = prb;
                } else {
                    cfgI->second = SumLogged(cfgI->second, prb);
                }
#ifdef ERR_DEBUG
                cerr <<  cfg << "\t" << cfgMap[cfg] << endl;
#endif
            }
        }
    }

    snpPair.m_ProbCfgs.reserve(cfgMap.size());
    for (TProbCfgMap::iterator cfgI = cfgMap.begin();  cfgI != cfgMap.end();
         ++cfgI) {
        //cerr << "\t" << cfgI->first << endl;
        snpPair.m_ProbCfgs.push_back(*cfgI);
    }

    /*    
    { unsigned int n = qDat.GetObservedCfg().N(); //debug
        T2LocusConfig cfg;
        cfg[e00] = n;
        TProbCfgs::iterator cfgI = snpPair.m_ProbCfgs.find(cfg);
        if (cfgI != snpPair.m_ProbCfgs.end() &&
            cfgI->second >= maxProb/10.) {
            cerr << "yo" << endl;
            snpPair.m_ProbCfgs.clear();
            snpPair.m_ProbCfgs[cfg] = log(1);
            //return;
        }
    }
    */

    //cerr << qDat.GetObservedCfg() << " yields " << snpPair.m_ProbCfgs.size() << " probcfgs." << endl;
}

/*---------------------------------------------------------------------------*/
// PRE : SNP iterator created
// POST: all relevant pairs of SNPs stored in m_SNPpairs
void CPIIMmodel::x_CacheSNPPairs(void) {
    static bool s_AlreadyWarned = false;
    /*
#ifdef LOAD_DEBUG
    cerr << "Caching SNP pairs.." << endl;
    cerr << "press enter to continue";
    char wait;
    cin >> wait;

    int c =0;
    double ave= 0;
    for (Csnp_CI i(GetAssmIterator());  i.good();  ++i) {
        ++c;
        *i;
        ave += i->GetCol().size();
        if (c % 1000 == 0) {
            cerr << c << "\t" << ave/1000 << "..\r";
            ave = 0;
            if (c == 200000) {
                exit(0);
            }
        }
    }
    cerr << "press enter to continue";
    cin >> wait;
#endif
    */
    /*
    //debug for picking random non-redundant pairs
    vector<unsigned int> snpS;
    gsl::Crng  rng(time(NULL));
    { unsigned int c = 0;
    for (Csnp_CI i(GetSNPIterator());  i.good();  ++i) {
        snpS.push_back(c++);
    }
    random_shuffle(snpS.begin(), snpS.end(), rng);
    }
    for (unsigned int a=0; a<snpS.size();  a += 2) {
        {
            Csnp_CI i(GetSNPIterator()+a);
            Csnp_CI j(GetSNPIterator()+(a+1));
    //debug end
    */
#ifdef DUMPIT
    ofstream dumpit("dumpit.txt");
#endif
    double lc = 0;
    double lcWithin = 0;
    unsigned int withinCnt = 0, betweenCnt = 0;
    unsigned int totCfgs = 0;

    /*
    for (Csnp_CI i(GetAssmIterator());  i.good();  ++i) {
        for (Csnp_CI j(i+1);  j.good()  &&  i.ChrOverlapsWith(j)  &&  j.GetPos() < i.GetPos()+40000;  ++j) {
    */
    for (CAssm_CI i(GetAssmIterator());  i.good();  i += sm_SiteIncrement) {
        //GetDepth includes Ns and gaps, which are not included in the Col
        if (i.GetDepth() <= 1  ||  i->GetCol().size() <= 1) {
            continue;
        }
        /*
        for (CAssm_CI j(i+11);  j.good()  &&  i.ChrOverlapsWith(j)  &&
                 j.GetPos() < i.GetPos()+40000;  ++j) {
        */
        CAssmOverlap_CI j(i);
        //need to skip ahead by least one so i != j
        do { ++j; } while(j.good()  &&
                          j.GetPos() < i.GetPos() + m_MinPairDist);

        for ( ;  j.good()  &&  j.GetPos() < i.GetPos()+40000;  ++j) {
            //GetDepth includes Ns and gaps, which are not included in the Col
            if (j.GetDepth() <= 1  ||  j->GetCol().size() <= 1) {
                continue;
            }
            /*
            if (j.OverlapsWith(i)) {
                continue;
            }
            */
            /*
            cerr << i.GetPos() << "(" << i.x_GetChrRangeStart() << "-" << i.x_GetChrRangeStop() << ")"
                 << "\t" << j.GetPos() << "(" << j.x_GetChrRangeStart() << "-" << j.x_GetChrRangeStop() << ")"<< endl;//DEBUG
            */
            SSNPPair snpPair;
            if (m_TrustObservations) {
                CQualCfg qDat(i->GetCol(), j->GetCol());
                snpPair.m_Dist = j.GetPos() - i.GetPos();
                snpPair.m_ProbCfgs.push_back(TProbCfgs::value_type
                                             (qDat.GetObservedCfg(), log(1)));
            } else {
                x_CalcObsProbCfgs(i, j, snpPair);
            }
            if (snpPair.m_ProbCfgs.empty()) {
#ifdef ERR_DEBUG
                cerr << "Skipping: " << i.GetPos() << "\t" << j.GetPos() << endl;
#endif
                /*
                cerr << "Skipping:" << endl
                     << i.GetPos() << "\t" << *i << endl
                     << j.GetPos() << "\t" << *j << endl;
                */
                continue;
            }
            /*
            CQualCfg qDat(i->GetCol(), j->GetCol());
            snpPair.m_ProbCfgs[qDat.GetObservedCfg()] = 0;
            */
            const T2LocusConfig &oneCfg = snpPair.m_ProbCfgs.begin()->first;
            if (oneCfg.N() > Get2LocusProbs().GetMaxDepth()) {
                //should probably issue better warning
                if (!s_AlreadyWarned) {
                    s_AlreadyWarned = true;
                    cerr << "Warning: data depth exceeds the maximum depth in the likelihood table at least once." << endl;
                }
                continue;
            }

            if (oneCfg.Nmissing() == oneCfg.N()) {
                //figure these guys give us so little information that
                //it's not worth it
                continue;
            }

            if (i.RdOverlapsWith(j)) {
                lcWithin += ((double)oneCfg.N()-oneCfg.Nmissing())/
                    (oneCfg.N()-oneCfg.Nmissing()/2);
                ++withinCnt;
            } else {
                lc += ((double)oneCfg.N()-oneCfg.Nmissing())/
                    (oneCfg.N()-oneCfg.Nmissing()/2);
                ++betweenCnt;
            }

            totCfgs += snpPair.m_ProbCfgs.size();
#ifdef LOAD_DEBUG2
            cerr << i.GetPos() << "\t" << snpPair.m_ProbCfgs.size() << endl;
#endif
#ifdef DUMPIT
            dumpit << i.GetPos() << "," << j.GetPos();
            for (unsigned int pcI = 0;  pcI < snpPair.m_ProbCfgs.size();  ++pcI) {
                dumpit << "\t" << snpPair.m_ProbCfgs[pcI].first << "-" << snpPair.m_ProbCfgs[pcI].second;
            }
            dumpit << endl;
#endif
            m_SNPpairs->push_back(snpPair);
        }
#ifdef MONITOR_LOAD
        //if (i.GetPos() % 10  == 0) {
        cerr << "processed L pos: " << i.GetPos() << " (" << m_SNPpairs->size() << " pairs, " << totCfgs << " cfgs so far)\r";
            //}
#endif
    }
    //cerr << "heck ya" << endl;
#ifdef LC_DEBUG
    cerr << "Linking coef (within): " << lcWithin/withinCnt << endl;
    cerr << "Linking coef ('tween): " << lc/betweenCnt << endl;
    cerr << "# of snp pairs: " << m_SNPpairs->size() << endl;
    cerr << "# of cfgs: " << totCfgs << endl;
#endif
#ifdef MONITOR_LOAD
    cerr << "Cached SNP pairs" << endl;
#endif
}

/*---------------------------------------------------------------------------*/
// PRE : parameters
// POST: loglikelihood of observing assembly w/ these parameters
double CPIIMmodel::CalcLogLik(const CParams &rawParams) const {
    if (rawParams.Get(eRho) < 0  || 
        rawParams.Get(eTheta) < 0  ||
        rawParams.Get(eGrowth) < 0) {
        throw range_error("Error: Negative parameters don't make any sense!");
    } else if (rawParams.Get(eTheta) > 1) {
        throw range_error("Error: Implausibly high theta.");
    }
    double loglik = 0;

    if (m_FreqSpec.empty()) { //must want to estimate recombination
        CParams p(rawParams);
#ifdef MONITOR_LIK
        int c = 0;
#endif
        for (TSNPPairs::const_iterator pairI = GetSNPPairs().begin();
             pairI != GetSNPPairs().end();  ++pairI) {
            //switch from per-site rho to full rho
            if (m_RecombModel == eCrossover) {
                p.Set(eRho) = rawParams.Get(eRho) * pairI->m_Dist;
            } else {
                //gene conversion model a la McVean, Frisse, Langley, etc.
                p.Set(eRho) = rawParams.Get(eRho) * 
                    2*rawParams.Get(eMeanTractLen) * 
                    (1 - exp(-(double)pairI->m_Dist/
                             rawParams.Get(eMeanTractLen)));
            }
            double pairLL = log(0);
            //double s=0;//DEBUG
            for (TProbCfgs::const_iterator cfgI = pairI->m_ProbCfgs.begin();
                 cfgI != pairI->m_ProbCfgs.end();  ++cfgI) {
                const T2LocusConfig &cfg = cfgI->first;
#ifdef ERR_DEBUG2
                cout << c << "\t" << exp(cfgI->second) << "\t" << cfg << endl;
#endif
                try {
                    pairLL = SumLogged(pairLL, cfgI->second +
                                       CalcPairLogLik(p, cfg));
                    //cerr << "yo: " << cfgI->second << "\t" << CalcPairLogLik(p, cfg) << endl;
                    //}
                } catch (const overflow_error &e) {
                    //really for debugging.. should delete if in production
                    ofstream dumpF("dump.xml");
                    SaveAssembly(dumpF);
                    throw;
                }
            }
            loglik += pairLL;
#ifdef MONITOR_LIK
            if (++c % 1000 == 0) {
               cerr << "processed pair " << c << "\r";
            }
#endif
        }
#ifdef MONITOR_LIK
        cerr << endl;
#endif
#ifdef ERR_DEBUG2
        cerr << loglik << endl;
        exit(0);
#endif
    } else { //want to estimate parameters from site frequency spectrum
        unique_ptr<CFreqSpectrum> fs;
        if (m_FreqSpec == "growth") {
            CFSGrowth *growthFS(new CFSGrowth);
            growthFS->SetParam(CFSGrowth::eTheta) = rawParams.Get(eTheta);
            growthFS->SetParam(CFSGrowth::eGrowthRate) =rawParams.Get(eGrowth);
            fs.reset(growthFS);
        } else {
            //fs.reset(new CFSTest);
            CFSNeutral *neutralFS(new CFSNeutral);
            neutralFS->SetParam(CFSNeutral::eTheta) = rawParams.Get(eTheta);
            if (neutralFS->ParamsOutsideBounds()) {
                throw range_error("");
            }
            fs.reset(neutralFS);
        }
        for (CAssm_CI i(GetAssmIterator());  i.good();  ++i) {
            loglik += log(i->GetPrObservedFS(*fs, m_TrustObservations));
        }
    }

    return loglik;
}

/*---------------------------------------------------------------------------*/
// PRE : parameters
// POST: loglikelihood of observing assembly w/ these parameters
void CPIIMmodel::CalcLogLikSurface(const CParams &rawParams,
                                   TLikSurface &loglikSurface) const {
    static bool s_AlreadyWarned = false;
    if (rawParams.Get(eRho) < 0  || 
        rawParams.Get(eTheta) < 0  ||
        rawParams.Get(eGrowth) < 0) {
        throw range_error("Error: Negative parameters don't make any sense!");
    } else if (rawParams.Get(eTheta) > 1) {
        throw range_error("Error: Implausibly high theta.");
    }

    unsigned int numGridPoints = 50;
    for (unsigned int rI = 0;  rI < numGridPoints;  ++rI) {
        double s = 7;//scaling factor to focus on sampling lower vals
        double maxRho = 0.5;
        CParams p(rawParams);
        p.Set(eRho, (maxRho/exp((numGridPoints-1)/s)) * exp(rI/s));
        loglikSurface.push_back(TLikSurface::value_type(p, log(1)));
    }

    for (CAssm_CI i(GetAssmIterator());  i.good();  i += sm_SiteIncrement) {
        //GetDepth includes Ns and gaps, which are not included in the Col
        if (i.GetDepth() <= 1  ||  i->GetCol().size() <= 1) {
            continue;
        }
        CAssmOverlap_CI j(i);
        do { ++j; } while(j.good()  && //skip ahead by least one so i != j
                          j.GetPos() < i.GetPos() + m_MinPairDist);

        for ( ;  j.good()  &&  j.GetPos() < i.GetPos()+40000;  ++j) {
            //GetDepth includes Ns and gaps, which are not included in the Col
            if (j.GetDepth() <= 1  ||  j->GetCol().size() <= 1) {
                continue;
            }
            SSNPPair snpPair;
            if (m_TrustObservations) {
                CQualCfg qDat(i->GetCol(), j->GetCol());
                snpPair.m_Dist = j.GetPos() - i.GetPos();
                snpPair.m_ProbCfgs.push_back(TProbCfgs::value_type
                                             (qDat.GetObservedCfg(), log(1)));
            } else {
                x_CalcObsProbCfgs(i, j, snpPair);
            }
            if (snpPair.m_ProbCfgs.empty()) {
                continue;
            }

            const T2LocusConfig &oneCfg = snpPair.m_ProbCfgs.begin()->first;
            if (oneCfg.N() > 40) {
                if (!s_AlreadyWarned) {
                    s_AlreadyWarned = true;
                    cerr << "Warning: data depth exceeds the maximum depth in the likelihood table at least once." << endl;
                }
                continue;
            }
            if (oneCfg.Nmissing() == oneCfg.N()) {
                //figure these guys give us so little information that
                //it's not worth it
                continue;
            }


            for (unsigned int rI = 0;  rI < numGridPoints;  ++rI) {
                CParams p(rawParams);
                //switch from per-site rho to full rho
                if (m_RecombModel == eCrossover) {
                    p.Set(eRho) = loglikSurface[rI].first.Get(eRho) *
                        snpPair.m_Dist;
                } else {
                    //gene conversion model a la McVean, Frisse, Langley, etc.
                    p.Set(eRho) = loglikSurface[rI].first.Get(eRho) * 
                        2*loglikSurface[rI].first.Get(eMeanTractLen) * 
                        (1 - exp(-(double)snpPair.m_Dist/
                                 loglikSurface[rI].first.Get(eMeanTractLen)));
                }

                double pairLL = log(0);
                for (TProbCfgs::const_iterator cfgI =
                         snpPair.m_ProbCfgs.begin();
                     cfgI != snpPair.m_ProbCfgs.end();  ++cfgI) {
                    const T2LocusConfig &cfg = cfgI->first;
                    try {
                        pairLL = SumLogged(pairLL, cfgI->second +
                                           CalcPairLogLik(p, cfg));
                    } catch (const range_error &e) {
                        //exceeded max tabulated rho value
                        numGridPoints = rI;
                        //2nd param below is dummy since will only be shrinking
                        loglikSurface.resize(rI, loglikSurface[0]);
                        break;
                    }
                }
                if (rI < numGridPoints) { //check still good
                    loglikSurface[rI].second += pairLL;
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/

int Main(TOpts &opts) {
    if (opts["version"]) {
        cout << "PIIM Version 2.0.3-r" << SVN_REVISION << endl;
        return 0;
    }

    ifstream likfile(opts["lk"]);
    if (!likfile.good()) {
        throw runtime_error("Cannot open input file '" + opts["lk"] + "'");
    }

    C2LocusProbs::EWhichSites whichSites = CAssm_CI::eAll;
    CPIIMmodel::ERecombModel recombModel;
    if (opts["model"] == "gc") {
        recombModel = CPIIMmodel::eGeneConv;
    } else if (opts["model"] == "crossover") {
        recombModel = CPIIMmodel::eCrossover;
    } else {
        throw runtime_error("Unknown recombination model parameter ("+
                            opts["model"]+")");
    }
    CPIIMmodel model(likfile, recombModel, whichSites,opts["mindist"].AsInt());
    CPIIMmodel::sm_SmallEnough = opts["small"].AsDouble();
    CPIIMmodel::sm_SiteIncrement = opts["site-incr"].AsInt();
    CAssm_CI::SetQualThresholds(opts["bad-q"].AsInt(), opts["maxbad"].AsInt());
    if (opts["trust-obs"]) {
        CAssm_CI::SetQualThresholds(0, 0);
        model.SetTrustObs(true);
    }

    CParams p(eNone); //default params for any likelihood calculation
    p.Set(eTheta, (opts["theta"] ? opts["theta"].AsDouble() :
                   model.Get2LocusProbs().GetTableTheta()));
    p.Set(eRho, 0.001);
    p.Set(eMeanTractLen, 500);

    {
        ifstream infile(opts['i']);
        if (opts['i'] != "-"  &&  !infile.good()) {
            throw runtime_error("Cannot open input file '" +
                                opts['i'] + "'");
        }
        istream& in(opts['i'] == "-" ? cin : infile);
        
        model.LoadAssembly(in,
                           opts["start"] ? opts["start"].AsInt() : -1,
                           opts["stop"] ? opts["stop"].AsInt() : -1);
    }

    if (opts["fs-save"]) {
        ofstream out(opts["fs-save"]);
        for (CAssm_CI i(model.GetAssmIterator());  i.good();  ++i) {
            out << i->GetPos() << "\t";
            i->WriteFSDist(out);
            out << endl;
        }
    }
    if (opts["fs-load"]) {
        ifstream in(opts["fs-load"]);
        model.LoadFSDist(in);
    }

    if (opts["joint-est"]) {
        model.SetFreqSpec("neutral");
        TParamNames pNames;
        pNames.push_back(eTheta);
        CFindMLE mle(model, p, pNames, 0.00001, -1, false);
        p.Set(eTheta) = mle.GetMLParams().Get(eTheta);
    }

    if (opts["surface"]) {
        CPIIMmodel::TLikSurface surface;
        model.CalcLogLikSurface(p, surface);
        unsigned int maxPos = 0;
        for (unsigned int i = 0;  i < surface.size();  ++i) {
            if (surface[i].second > surface[maxPos].second) {
                maxPos = i;
            }
            if (opts["gridout"]) {
                cout << "ll: " << setprecision(5) << fixed
                     << surface[i].second << setprecision(5)
                     << resetiosflags(ios::fixed) << "\t"
                     << surface[i].first << endl;
            }
        }
        if (maxPos == surface.size()-1) {
            surface[maxPos].first.Set(eRho) = INFINITY;
        } else {
            //fit to cubic using 7 grid points (max +- 3)
            gsl::CVector coefs(4);
            gsl::CMatrix x(7, coefs.size());
            gsl::CVector y(x.nrow());
            int firstRho = max(0,(int)maxPos-(int)y.size()/2);
            int lastRho = firstRho + y.size()-1;
            for (int rhoI = firstRho;  rhoI <= lastRho;  ++rhoI) {
                for (unsigned int i = 0;  i < coefs.size(); ++i) {
                    gsl_matrix_set(x, rhoI-firstRho, i,
                                   pow(surface[rhoI].first.Get(eRho), (int)i));
                }
                gsl_vector_set(y, rhoI-firstRho, surface[rhoI].second);
            }
            gsl::CMultiFit::FitData(x,y,coefs);

            //calc concave-down zero of cubic
            double discr = 4*coefs.get(2)*coefs.get(2) -
                12*coefs.get(1) * coefs.get(3);
            double zero1 = (-2 * coefs.get(2) + sqrt(discr)) /(6 * coefs.get(3));
            double zero2 = (-2 * coefs.get(2) - sqrt(discr)) /(6 * coefs.get(3));
            double mle = max(0.0, (6*coefs.get(3)*zero1 + 2*coefs.get(2) < 0) ?
                             zero1 : zero2);
            if ((maxPos > 0  &&  mle < surface[maxPos-1].first.Get(eRho))  ||
                (maxPos < surface.size()-1  &&
                 mle > surface[maxPos+1].first.Get(eRho))) {
                cerr << "Warning: cubic mle (" << mle
                     << ") is outside the grid bounds ("
                     << surface[max(0,(int)(maxPos)-1)].first.Get(eRho) << "\t"
                     << surface[min((int)surface.size()-1,(int)maxPos+1)].
                    first.Get(eRho)
                     << " (probably means the likelihood surface is not completely smooth around the MLE); using gridded MLE instead" << endl;
            } else {
                surface[maxPos].first.Set(eRho) = mle;
            }
        }
        cout << surface[maxPos].first << endl;
        return 0;
    }

    if (opts["freqspec"]) {
        model.SetFreqSpec(opts["freqspec"]);
        TParamNames pNames;
        pNames.push_back(eTheta);
        if (opts["freqspec"] == "growth") {
            pNames.push_back(eGrowth);
        }
        CFindMLE mle(model, p, pNames, 0.00001, -1, false);
        cout << model.CalcLogLik(mle.GetMLParams()) << "\t"
             << mle.GetMLParams() << endl;
    } else {
        model.SetFreqSpec("");

        TParamNames pNames;
        //pNames.push_back(eTheta);
        pNames.push_back(eRho);
        
        CFindMLE mle(model, p, pNames, 0.0001, -1, false);
        CParams mlParams(mle.GetMLParams());
        //should we declare this an infinite estimate?
        if (mle.GetMaxExploredParams().Get(eRho) == mlParams.Get(eRho)) {
            mlParams.Set(eRho) = INFINITY;
        }
        cout << mlParams;
        if (opts["ci"]) {
            CParams paramUB(eNone), paramLB(eNone);
            CFindCI(model, mlParams, pNames, paramLB, paramUB, 0.001);
            cout << "\t" << paramLB.Get(eRho) << "\t" << paramUB.Get(eRho);
        }
        cout << endl;
    }

    return 0;
}

/*---------------------------------------------------------------------------*/
int main(int argc, char** argv) {
    try {
        TOpts opts;
        opts['i'].SetRequired(true).SetDesc("input XML filename");
        opts["lk"].SetRequired(true).
            SetDesc("file containing 2-locus likelihoods");
        opts["start"].SetDesc("start position on assembly");
        opts["stop"]. SetDesc("stop position on assembly");
        opts["fs-save"].SetDesc("save o|t f.s. distribution");
        opts["fs-load"].SetDesc("load o|t f.s. distribution");
        opts["gridout"].SetFlag();
        opts["model"].SetDesc("gc | crossover") = "gc";
        opts["small"].SetDesc("how small a probability will we still worry about?") = "1e-4";
        opts["maxbad"].SetDesc("max number of \"bad\" bases to allow in a pair (uses threshold set by -bad-q)") = "5";
        opts["bad-q"].SetDesc("quality scores below this level are considered \"bad\"") = "20";
        opts["mindist"].SetDesc("minimum distance between pairs of sites") = "10";
        opts["site-incr"].SetDesc("increment to shift when picking sites") = 
            "1";
        opts["trust-obs"].SetFlag().SetDesc("Trust observations completely & do not condition on q-scores.");
        opts["surface"].SetFlag().SetDesc("Calc likelihood surface in grid across rho in one pass.");
        opts["theta"].SetDesc("theta to use when estimating rho");
        opts["joint-est"].SetFlag().SetDesc("Estimate theta from FS, then use while estimating rho");
        opts["version"].SetFlag().SetOverrideRequired(true).SetDesc("Report program version");

        x_ProcessOptions(argc, argv, opts,
                         "Population Inference in Metagenomics (PIIM) -- "
                         "now, with recombination");
        return Main(opts);
    } catch (exception &e) {
        cerr << "Exception in PIIM:\n\t" << e.what() << endl;
        return 1;
    }
}
