/* $Id: util_disklist.hpp 156 2010-09-13 21:15:29Z pjohnson $
  ---------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2009. "Inference of Microbial
    Recombination Rates from Metagenomic Data."  PLoS Genetics (in
    press).
 
  Copyright 2009 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Template class to store record on disk via "ReadBinary" and
  "WriteBinary" methods with a STL list-like interface.  Actually can
  only handle a single iterator at a time & can't mix writing
  (push_back) and reading (iterator) -- but designed to look
  interchangable..
  ---------------------------------------------------------------------------
*/
#ifndef UTIL_DISKLIST_HPP
#define UTIL_DISKLIST_HPP

#include <cstdio>

using namespace std;

//---------------------------------------------------------------------------
// CFileBuffer class is rather fragile.. works for the time being, but
// be careful!
class CFileBuffer {
public:
    CFileBuffer(FILE *f, unsigned int sz) : m_File(f), m_BuffSize(sz) {
        m_Writing = false;
        m_BuffI = m_BuffEnd = m_Buff = NULL;
    }
    ~CFileBuffer(void) { if (m_Writing) {x_WriteBuffer();} free(m_Buff); }

    void rewind(void) {
        if (m_Writing) {
            x_WriteBuffer();
            m_Writing = false;
        }
        fflush(m_File);
        ::rewind(m_File);
        m_BuffI = m_BuffEnd;
    }

    unsigned long tell(void) const {
        return m_BuffI - m_Buff;    //maybe mod to be true tell, eh?
    }

    unsigned int read(char *dest, unsigned int bytes) {
        /*
        if (!m_Buff) {
            struct stat stats;
            fstat(m_Fd, &stats);
            m_Buff = (char*) mmap(NULL, stats.st_size, PROT_READ,
                                  MAP_SHARED, m_Fd, 0);
            if (m_Buff == MAP_FAILED) {
                throw runtime_error("mmap failed");
            }
            m_BuffI = m_Buff;
            m_BuffEnd = m_Buff + stats.st_size;
        }
        assert(m_BuffI + bytes <= m_BuffEnd);
        memcpy(dest, m_BuffI, bytes);
        m_BuffI += bytes;
        return bytes;
        */

        unsigned int bytesRead = min(bytes, (unsigned int)(m_BuffEnd-m_BuffI));
        memcpy(dest, m_BuffI, bytesRead);
        m_BuffI += bytesRead;
        while (bytesRead < bytes  &&  x_FillBuffer()) {
            unsigned int toRead = min(bytes-bytesRead,
                                      (unsigned int)(m_BuffEnd-m_BuffI));
            memcpy(dest+bytesRead, m_BuffI, toRead);
            m_BuffI += toRead;
            bytesRead += toRead;
        }
        if (bytesRead != bytes) {
            cerr << "Error: read read " << bytesRead << " instead of " << bytes
                 << endl;
            throw runtime_error("::read crunched");
        }
        return bytesRead;
    }

    unsigned int write(const char *src, unsigned int bytes) {
        if (!m_Writing) { //if we're not already writing, reset ptr to begin
            m_Writing = true;
            if (!m_Buff) {
                m_Buff = (char*) malloc(m_BuffSize);
                m_BuffEnd = m_Buff + m_BuffSize;
            }
            m_BuffI = m_Buff;
        }
        unsigned int bytesWrit = min(bytes, (unsigned int)(m_BuffEnd-m_BuffI));
        memcpy(m_BuffI, src, bytesWrit);
        m_BuffI += bytesWrit;
        while (bytesWrit < bytes  &&  x_WriteBuffer()) {
            unsigned int toWrit = min(bytes-bytesWrit,
                                      (unsigned int)(m_BuffEnd-m_BuffI));
            memcpy(m_BuffI, src+bytesWrit, toWrit);
            m_BuffI += toWrit;
            bytesWrit += toWrit;
        }
        return bytesWrit;
    }

private:
    bool x_FillBuffer(void) {
#ifdef DISKLIST_DBG
        cerr << "Filling buffer..";
#endif
        if (!m_Buff) {
            m_Buff = (char*) malloc(m_BuffSize);
        }
        m_BuffEnd = m_Buff + fread((void*)m_Buff, 1, m_BuffSize, m_File);
        m_BuffI = m_Buff;
#ifdef DISKLIST_DBG
        cerr << "done" << endl;
#endif
        return (m_BuffEnd - m_Buff > 0);
    }

    bool x_WriteBuffer() {
#ifdef DISKLIST_DBG
        cerr << "Writing buffer..";
#endif
        size_t bytes = fwrite((const void*)m_Buff, 1, m_BuffI-m_Buff, m_File);
        if ((int)bytes != m_BuffI-m_Buff) {
            cerr << "Error: x_WriteBuffer wrote " << bytes << " instead of " << m_BuffI-m_Buff
                 << endl;
            throw runtime_error("::write crunched");
        }
        m_BuffI = m_Buff;
#ifdef DISKLIST_DBG
        cerr << "done" << endl;
#endif
        return (bytes > 0);
    }

private:
    FILE *m_File;
    size_t m_BuffSize;
    bool m_Writing;
    char *m_Buff, *m_BuffI, *m_BuffEnd;
};


template <typename Tcontents, unsigned int kBuffSize>
class CDiskList {
public:
    typedef CDiskList<Tcontents, kBuffSize> TThisType;

    class const_iterator {
        friend class CDiskList;
    public:
        const_iterator(TThisType& dl) :
            m_DiskList(dl), m_CurRecord(0), m_Loaded(false) {}

        const Tcontents& operator* (void) {
            if (!m_Loaded) {
                m_DiskList.x_read(m_Record);
                m_Loaded = true;
            }
            return m_Record;
        }
        const Tcontents* operator-> (void) { return &(**this); }
        const_iterator& operator++ (void) {
            m_Loaded = false;
            ++m_CurRecord;
            return *this;
        }
        bool operator!= (const const_iterator &it) const {
            return it.m_CurRecord != m_CurRecord;
        }

    private:
        TThisType &m_DiskList;
        int m_CurRecord;
        bool m_Loaded;
        Tcontents m_Record;
    };

public:
    CDiskList(void) : m_File(tmpfile()), m_EndIt(*this),
                      m_Buffer(m_File, kBuffSize) {
        m_NumRecords = 0;
    }
    ~CDiskList(void) { fclose(m_File); }
    const const_iterator begin(void) const {
        const_cast<TThisType*>(this)->x_rewind();
        return const_iterator(const_cast<TThisType&>(*this));
    }
    const const_iterator& end(void) const { return m_EndIt; }
    unsigned int size(void) const { return m_NumRecords; }
    void push_back(const Tcontents& c) {
        c.WriteBinary(m_Buffer);
        ++m_NumRecords;
        m_EndIt.m_CurRecord = m_NumRecords;
    }

protected:
    void x_rewind(void) {
        m_Buffer.rewind();
    }
    void x_read(Tcontents& record) {
        record.ReadBinary(m_Buffer);
    }
    
private:
    FILE *m_File;
    const_iterator m_EndIt;
    unsigned int m_NumRecords;
    CFileBuffer m_Buffer;
};

//---------------------------------------------------------------------------


#endif /*ifndef UTIL_DISKLIST_HPP*/
